#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Libraries
import re
import pandas as pd

# Modules
import datamodel
from tools import Logger


questions = [
    (r"What is (.*)", lambda matched_str, **kwargs: matched_what(matched_str, **kwargs)),

    (r"Is (.*) allocated", lambda matched_str, **kwargs: matched_allocated(matched_str, **kwargs)),
]


def find_question(string, **kwargs):
    """Entry point from jarvis.py"""
    out = lookup_table(string, questions, **kwargs)
    return out


def lookup_table(strings, questions_table, **kwargs):
    """Method to match regex and corresponding method from [questions]"""
    answer_list = []
    for regex, method in questions_table:
        for i in strings:
            result = re.findall(regex, i, re.MULTILINE)
            if result:
                answer = method(result, **kwargs)
                if answer:
                    answer_list.append(answer)
    return answer_list


def matched_what(question_str, **kwargs):
    """Get the 'what' declaration """
    object_str = question_str[0].strip()
    wanted_object = check_get_object(object_str, **kwargs)

    if wanted_object:
        object_info = get_object_info(wanted_object, **kwargs)
        if object_info:
            return object_info


def matched_allocated(object_str, **kwargs):
    """From object_str, get object then check if allocated and returns allocation's list"""
    object_info = ""
    object_str = object_str[0]
    xml_function_name_list = get_objects_names(kwargs['xml_function_list'])
    xml_state_name_list = get_objects_names(kwargs['xml_state_list'])
    whole_objects_name_list = [*xml_function_name_list, *xml_state_name_list]
    if not any(s == object_str for s in whole_objects_name_list):
        Logger.set_warning(__name__,
                          f"{object_str} is not a function nor a state")
    else:
        result_function = any(s == object_str for s in xml_function_name_list)
        resul_state = any(s == object_str for s in xml_state_name_list)
        result = [result_function, False, resul_state,  False, False]
        wanted_object = match_object(object_str, result, **kwargs)
        if wanted_object:
            allocation_list = get_allocation_object(wanted_object, kwargs['xml_fun_elem_list'])
            if allocation_list:
                object_info += f'"{wanted_object.name}" is allocated to ' \
                               + ", ".join([elem.name for elem in allocation_list])
                return object_info

    return

xml_str_lists = ['xml_function_list', 'xml_data_list', 'xml_state_list', 'xml_fun_elem_list',
                    'xml_transition_list', 'xml_fun_inter_list', 'xml_phy_elem_list',
                    'xml_phy_inter_list', 'xml_attribute_list', 'xml_view_list', 'xml_type_list']

def get_objects_name_lists(**kwargs):
    """Returns lists of objects with their names depending on kwargs"""
    whole_objects_name_list = [[] for _ in range(11)]
    for i in range(11):
        if kwargs.get(xml_str_lists[i], False):
            whole_objects_name_list[i] = get_objects_names(kwargs[xml_str_lists[i]])

    return whole_objects_name_list


# TODO: Use this method within xxxxxxx_orchestrator.py
def check_get_object(object_str, **kwargs):
    """
    Returns the desired object from object's string
    Args:
        object_str ([object_string]): list of object's name from cell
        **kwargs: xml lists

    Returns:
        wanted_object : Function/State/Data/Fun_Elem/Transition/Fun_Inter
    """
    whole_objects_name_list = get_objects_name_lists(**kwargs)
    if not any(object_str in s for s in whole_objects_name_list):
        return None
    else:
        result = [False] * 11
        for i in range(11):
            result[i] = any(a == object_str for a in whole_objects_name_list[i])

        wanted_object = match_object(object_str, result, xml_str_lists=xml_str_lists, **kwargs)
        return wanted_object


def match_object(object_str, result, xml_str_lists=None, **kwargs):
    """Returns wanted_object from object_str and result matched from name lists"""
    # Because match_object() called within match_allocated() TBC/TBT if match_allocated()
    # still needed
    if not xml_str_lists:
        xml_str_lists = ['xml_function_list', 'xml_data_list', 'xml_state_list',
                         'xml_fun_elem_list', 'xml_transition_list', 'xml_fun_inter_list',
                         'xml_phy_elem_list', 'xml_phy_inter_list', 'xml_attribute_list',
                         'xml_view_list', 'xml_type_list']
    for i in range(11):
        if result[i]:
            for obj in kwargs[xml_str_lists[i]]:
                if object_str == obj.name:
                    return obj
                try:
                    if object_str == obj.alias:
                        return obj
                except AttributeError:
                    # To avoid error when there is no alias for the object
                    pass
    return None


def get_object_info(wanted_object, **kwargs):
    """From wanted_object and all lists, returns a dict with info"""
    object_info = {'Name': str(wanted_object.name),
                   'Object Class': str(type(wanted_object).__name__),
                   'Type': str(wanted_object.type), 'Id': str(wanted_object.id)}

    try:
        # Only Data() does not have these attributes and Transition() does not have parent
        if any(wanted_object.alias):
            object_info['Alias'] = str(wanted_object.alias)
        if wanted_object.parent is not None:
            object_info['Parent'] = str(wanted_object.parent.name)
    except AttributeError:
        # To avoid error when there is no such attribute for the object
        pass
    if object_info['Object Class'] == 'Function':
        get_function_info(wanted_object, object_info, **kwargs)
    elif object_info['Object Class'] == 'FunctionalElement':
        get_fun_elem_info(wanted_object, object_info, **kwargs)
    elif object_info['Object Class'] == 'State':
        if any(wanted_object.child_list):
            object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_state_list'])
    elif object_info['Object Class'] == 'Transition':
        get_transition_info(wanted_object, kwargs['xml_state_list'], object_info)
    elif object_info['Object Class'] == 'Data':
        get_data_info(wanted_object, object_info, **kwargs)

    if object_info:
        return object_info

    return


def get_function_info(wanted_object, object_info, **kwargs):
    """Get Function info"""
    if any(wanted_object.child_list):
        object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_function_list'])
    if wanted_object.input_role is not None:
        object_info['Input Role'] = str(wanted_object.input_role)
    if wanted_object.operand is not None:
        object_info['Operand'] = str(wanted_object.operand)

    object_info['Consumption List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_consumer_function_list'])
    object_info['Production List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_producer_function_list'])
    return object_info


def get_consumes_produces_info(wanted_object, relationship_list):
    """Get consumes/produces info"""
    object_relationship = set()
    for elem in relationship_list:
        if elem[1].id == wanted_object.id:
            object_relationship.add(elem[0])
        if elem[0] == wanted_object.name:
            object_relationship.add(elem[1].name)
    if object_relationship:
        return object_relationship


def get_child_name_list(parent_object, object_list):
    """Get child's name list"""
    child_list = set()
    for child in object_list:
        if child in parent_object.child_list:
            child_list.add((child.name, "Child"))
    return list(child_list)


def get_allocated_object_name_list(wanted_object, object_list):
    """From Functional Element object and list (State/Function) get the allocated objects"""
    allocation_list = set()
    for allocated_object in object_list:
        if any(s == allocated_object.id for s in wanted_object.allocated_state_list):
            allocation_list.add(allocated_object.name)
        if any(s == allocated_object.id for s in wanted_object.allocated_function_list):
            allocation_list.add(allocated_object.name)
    return allocation_list


def get_transition_info(wanted_object, state_list, object_info):
    """Get transition's info"""
    for state in state_list:
        if state.id == wanted_object.source:
            object_info['Source'] = str(state.name)
        if state.id == wanted_object.destination:
            object_info['Destination'] = str(state.name)
    if wanted_object.condition_list:
        object_info['Condition List'] = wanted_object.condition_list
    return object_info


def get_fun_elem_info(wanted_object, object_info, **kwargs):
    """Get functional element info, i.e. child_list, allocated Function/State"""
    if any(wanted_object.child_list):
        object_info['Child List'] = get_child_name_list(wanted_object, kwargs['xml_fun_elem_list'])
    if any(wanted_object.allocated_state_list):
        object_info['Allocated State List'] = \
            get_allocated_object_name_list(wanted_object, kwargs['xml_state_list'])
    if any(wanted_object.allocated_function_list):
        object_info['Allocated Function List'] = \
            get_allocated_object_name_list(wanted_object, kwargs['xml_function_list'])
    return object_info


def get_data_info(wanted_object, object_info, **kwargs):
    """Get what consumes/produces for a specific Data object"""
    pred_list = set()
    object_info['Consumer List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_consumer_function_list'])
    object_info['Producer List'] = \
        get_consumes_produces_info(wanted_object, kwargs['xml_producer_function_list'])
    if any(wanted_object.predecessor_list):
        for pred in wanted_object.predecessor_list:
            pred_list.add(pred.name)
    object_info['Predecessor List'] = pred_list
    return object_info


def get_allocation_object(wanted_object, object_list):
    """Get current allocation for an object"""
    allocation_set = set()
    object_type = get_object_type(wanted_object)

    if object_type == 'function':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.allocated_function_list):
                allocation_set.add(fun_elem)
    elif object_type == 'state':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.allocated_state_list):
                allocation_set.add(fun_elem)
    elif object_type == 'data':
        for fun_inter in object_list:
            if any(s == wanted_object.id for s in fun_inter.allocated_data_list):
                allocation_set.add(fun_inter)
    elif object_type == 'Functional interface':
        for fun_elem in object_list:
            if any(s == wanted_object.id for s in fun_elem.exposed_interface_list):
                allocation_set.add(fun_elem)
    elif object_type == 'Functional element':
        for function in object_list:
            if any(s == function.id for s in wanted_object.allocated_function_list):
                allocation_set.add(function)
    if allocation_set:
        return allocation_set

    return


def switch_objects_lists(type_list_str, wanted_object, object_type, **kwargs):
    """Switch depending on list's type and object's type """
    if object_type in ("state", "function", "Functional element"):
        switch_type_list = {
            "input": switch_in,
            "output": switch_out,
            "child": switch_child,
            "function": switch_state_function,
            "transition": switch_state_transition,
            "interface": switch_fun_elem_interface,
        }
        if object_type == "state" and type_list_str in ("input", "output"):
            return case_no_list(wanted_object, object_type, **kwargs)
        elif object_type != "state" and type_list_str in ("function", "transition"):
            return case_no_list(wanted_object, object_type, **kwargs)
        elif object_type != "Functional element" and type_list_str == "interface":
            return case_no_list(wanted_object, object_type, **kwargs)
        else:
            type_list = switch_type_list.get(type_list_str, case_no_list)
            return type_list(wanted_object, object_type, **kwargs)

    elif object_type == "Functional interface" and type_list_str == "data":
        return switch_data(wanted_object, object_type, **kwargs)

    else:
        return case_no_list(wanted_object, object_type, **kwargs)


def switch_in(wanted_object, _=None, **kwargs):
    """Case 'list input Function/Functional ELement' """
    input_list = get_input_or_output_fun_and_fun_elem(wanted_object, direction='input', **kwargs)
    if wanted_object.derived:
        input_list.append(*get_input_or_output_fun_and_fun_elem(wanted_object.derived,
                                                                direction='input', **kwargs))
    if input_list:
        input_dict = {'title': f"Input list for {wanted_object.name}:",
                      'data': input_list,
                      'columns': ["Data name", "Producer"]}
        return input_dict


def switch_out(wanted_object, _=None, **kwargs):
    """Case 'list output Function/Functional ELement' """
    output_list = get_input_or_output_fun_and_fun_elem(wanted_object, direction='output', **kwargs)
    if wanted_object.derived:
        output_list.append(*get_input_or_output_fun_and_fun_elem(wanted_object.derived,
                                                                 direction='output', **kwargs))
    if output_list:
        output_dict = {'title': f"Output list for {wanted_object.name}:",
                       'data': output_list,
                       'columns': ["Data name", "Consumer"]}
        return output_dict


def switch_child(wanted_object, object_type, **kwargs):
    """Case 'list child Function/State/Functional ELement' """
    child_list = None
    if object_type == "function":
        child_list = get_child_name_list(wanted_object, kwargs['xml_function_list'])
        if wanted_object.derived:
            child_list += [e for e in get_child_name_list(wanted_object.derived,
                                                          kwargs['xml_function_list'])]
    elif object_type == "state":
        child_list = get_child_name_list(wanted_object, kwargs['xml_state_list'])

    elif object_type == "Functional element":
        child_list = get_child_name_list(wanted_object, kwargs['xml_fun_elem_list'])

        child_list.extend(get_fun_elem_function_state_allocation(wanted_object,
                                                                 kwargs['xml_function_list'],
                                                                 kwargs['xml_state_list']))
        if wanted_object.derived:
            child_list += [e for e in get_child_name_list(wanted_object.derived,
                                                          kwargs['xml_fun_elem_list'])]
            child_list += [e for e in get_fun_elem_function_state_allocation(
                wanted_object.derived, kwargs['xml_function_list'], kwargs['xml_state_list'])]

    if child_list:
        child_dict = {'title': f"Child list for {wanted_object.name}:",
                      'data': list(tuple(sorted(child_list))),
                      'columns': ["Object's name", "Relationship's type"]}
        return child_dict


def get_fun_elem_function_state_allocation(wanted_object, xml_function_list, xml_state_list):
    """Returns a list for allocations with:
    [(function.name, "Function allocation"), (state.name, "State allocation"), ...]
    """
    allocation_list = []
    for allocated_fun in wanted_object.allocated_function_list:
        for fun in xml_function_list:
            if fun.id == allocated_fun:
                allocation_list.append((fun.name, "Function allocation"))
    for allocated_state in wanted_object.allocated_state_list:
        for state in xml_state_list:
            if state.id == allocated_state:
                allocation_list.append((state.name, "State allocation"))

    return allocation_list


def switch_state_function(wanted_object, _, **kwargs):
    """Case 'list function State' """
    function_list = []
    for allocated_fun in wanted_object.allocated_function_list:
        for fun in kwargs['xml_function_list']:
            if fun.id == allocated_fun:
                function_list.append((fun.name, "Function allocation"))

    if function_list:
        function_dict = {'title': f"Function list for {wanted_object.name}:",
                         'data': list(tuple(sorted(function_list))),
                         'columns': ["Object's name", "Relationship's type"]}
        return function_dict


def switch_state_transition(wanted_object, _, **kwargs):
    """Case 'list transition State' """
    transition_list = []
    for transition in kwargs['xml_transition_list']:
        if wanted_object.id == transition.source:
            for state in kwargs['xml_state_list']:
                if transition.destination == state.id:
                    transition_list.append({
                        'Transition name': transition.name,
                        'Source state': wanted_object.name,
                        'Destination state': state.name,
                        'Condition(s)': transition.condition_list
                    })
        elif wanted_object.id == transition.destination:
            for state in kwargs['xml_state_list']:
                if transition.source == state.id:
                    transition_list.append({
                        'Transition name': transition.name,
                        'Source state': state.name,
                        'Destination state': wanted_object.name,
                        'Condition(s)': transition.condition_list
                    })

    if transition_list:
        transition_dict = {'title': f"Transition list for {wanted_object.name}:",
                           'data': transition_list}
        return transition_dict


def switch_fun_elem_interface(wanted_object, _, **kwargs):
    """Case for 'list interface Functional element'"""
    id_list = wanted_object.exposed_interface_list
    main_fun_elem_child_list, _ = get_children(wanted_object)
    if wanted_object.derived:
        id_list = id_list.union(wanted_object.derived.exposed_interface_list)
        main_fun_elem_child_list = main_fun_elem_child_list.union(
            get_children(wanted_object.derived)[0])

    fun_inter_list = get_objects_from_id_list(id_list,
                                              kwargs['xml_fun_inter_list'])

    if not fun_inter_list:
        return "Not any exposed interface for {!s}".format(wanted_object.name)

    exposing_fun_elem = set()
    for interface in fun_inter_list:
        for fun_elem in kwargs['xml_fun_elem_list']:
            if fun_elem not in main_fun_elem_child_list and \
                    interface.id in fun_elem.exposed_interface_list and \
                    check_not_family(fun_elem, wanted_object):
                exposing_fun_elem.add((interface, fun_elem))

    interface_list = set()
    for k in exposing_fun_elem:
        child_list, _ = get_children(k[1])
        child_list.remove(k[1])
        if child_list:
            check = True
            for p in exposing_fun_elem:
                if p[0] == k[0] and any(a == p[1] for a in child_list):
                    check = False
                    break
            if check:
                interface_list.add((k[0].name, k[1].name))
        else:
            interface_list.add((k[0].name, k[1].name))
    if interface_list:
        interface_dict = {'title': f"Interface list for {wanted_object.name}:",
                          'data': list(tuple(sorted(interface_list))),
                          'columns': ["Interface ", "Last connected functional element"]}
        return interface_dict


def switch_data(wanted_object, _, **kwargs):
    """Case for 'list data Functional Interface' """
    data_list = []
    fun_elem_exposing = get_allocation_object(wanted_object, kwargs['xml_fun_elem_list'])
    if wanted_object.derived:
        derived_fun_elem_exposing = get_allocation_object(wanted_object.derived,
                                                          kwargs['xml_fun_elem_list'])
        if derived_fun_elem_exposing and fun_elem_exposing:
            fun_elem_exposing = fun_elem_exposing.union(derived_fun_elem_exposing)

    if not fun_elem_exposing:
        return f"Not any functional element exosing {wanted_object.name}"
    last_fun_elem_exposing = [check_latest(j, fun_elem_exposing) for j in fun_elem_exposing
                              if check_latest(j, fun_elem_exposing)]

    for allocated_id in wanted_object.allocated_data_list:
        for data in kwargs['xml_data_list']:
            if allocated_id == data.id:
                data_list.append(get_latest_obj_interface(data, last_fun_elem_exposing, **kwargs))

    if wanted_object.derived:
        for allocated_id in wanted_object.derived.allocated_data_list:
            for data in kwargs['xml_data_list']:
                if allocated_id == data.id:
                    data_list.append(
                        get_latest_obj_interface(data, last_fun_elem_exposing, **kwargs))

    if data_list:
        data_dict = {'title': f"Data list for {wanted_object.name}:",
                     'data': data_list}
        return data_dict


def case_no_list(wanted_object, object_type, _):
    """Case when there is incompatible list's type with object's type """
    return f"No list available for object '{wanted_object.name}' " \
           f"of type '{object_type.capitalize()}', possible lists are:\n" \
           f"- List child [Function/State/Functional element]\n" \
           f"- List input/output [Function/Functional element]\n" \
           f"- List function/transition [State]\n" \
           f"- List interface [Functional element]\n" \
           f"- List data [Functional interface]"


def get_latest_obj_interface(data, last_fun_elem_exposing, **kwargs):
    """For a data, find last producer and consumer if they are allocated to last fun_elem
    exposing the functional interface asked"""
    data_dict = {'Data': data.name,
                 'Last consumer Function(s)': [],
                 'Last consumer Functional element(s)': [],
                 'Last producer Function(s)': [],
                 'Last producer Functional element(s)': []}
    for prod in kwargs['xml_producer_function_list']:
        if prod[0] == data.name and \
                check_latest(prod[1], kwargs['xml_function_list']) == prod[1].name:
            for cons in kwargs['xml_consumer_function_list']:
                cons_last_fun_elem = None
                prod_last_fun_elem = None
                if cons[0] == prod[0] and \
                        check_latest(cons[1], kwargs['xml_function_list']) == cons[1].name:
                    cons_fun_elem_list = get_allocation_object(cons[1], kwargs['xml_fun_elem_list'])
                    if cons_fun_elem_list:
                        for fun_elem in cons_fun_elem_list:
                            if fun_elem.name in last_fun_elem_exposing:
                                cons_last_fun_elem = fun_elem
                    prod_fun_elem_list = get_allocation_object(prod[1], kwargs['xml_fun_elem_list'])
                    if prod_fun_elem_list:
                        for fun_elem in prod_fun_elem_list:
                            if fun_elem.name in last_fun_elem_exposing:
                                prod_last_fun_elem = fun_elem

                if not cons_last_fun_elem == prod_last_fun_elem:
                    if not any(c == prod[1].name for c in data_dict['Last producer Function(s)']):
                        if prod_last_fun_elem and \
                                prod[1].id in prod_last_fun_elem.allocated_function_list:
                            data_dict['Last producer Function(s)'].append(prod[1].name)
                            data_dict['Last producer Functional element(s)'].append(
                                prod_last_fun_elem.name)
                    if not any(c == cons[1].name for c in data_dict['Last consumer Function(s)']):
                        if cons_last_fun_elem and \
                                cons[1].id in cons_last_fun_elem.allocated_function_list:
                            data_dict['Last consumer Function(s)'].append(cons[1].name)
                            data_dict['Last consumer Functional element(s)'].append(
                                cons_last_fun_elem.name)

    return data_dict


def check_latest(wanted_object, check_list):
    """Checks and returns latest object = last one decomposed"""
    if not wanted_object.child_list:
        return wanted_object.name
    else:
        check_child = [c for c in wanted_object.child_list]
        if not any(j in check_child for j in check_list):
            return wanted_object.name


def get_object_list(object_str, **kwargs):
    """
    Gets lists from object_str : i.e. Input/Output/Child/Data
    Args:
        object_str ([object_string]): list of object's name from cell
        **kwargs: all xml lists

    Returns:
        answer_list: [Input_list, Output_list, Child_list, Data_list]
    """
    answer_list = []
    for elem in object_str:
        wanted_object = check_get_object(elem[1], **kwargs)
        if wanted_object is None:
            Logger.set_error(__name__,
                            f"Object '{elem[1]}' does not exist")
        else:
            object_type = get_object_type(wanted_object)
            wanted_list = switch_objects_lists(elem[0], wanted_object, object_type, **kwargs)
            if isinstance(wanted_list, (list, dict)):
                answer_list.append(wanted_list)
            elif isinstance(wanted_list, str):
                # Single display (not related to logging)
                print(wanted_list)
            else:
                Logger.set_info(__name__,
                               f"Nothing to display for {elem[0]} list of '{wanted_object.name}'")

    return answer_list


def get_input_or_output_fun_and_fun_elem(wanted_object, direction='input', unmerged=False,
                                         **kwargs):
    """
    Gets inputs/outputs for object (Function or Functional Element):
    i.e. what is consumed/produces by wanted object or object allocated functions.
    Args:
        wanted_object: current object
        direction (str: default=input): i.e. input or output asked
        unmerged (bool: default=false): used by functional elements
        **kwargs: all xml lists/sets

    Returns:
        input or output list
    """
    object_type = get_object_type(wanted_object)
    if object_type == "Functional element":
        in_or_out_list = []
        allocated_fun_list = set()
        for fun in wanted_object.allocated_function_list:
            for xml_fun in kwargs['xml_function_list']:
                if fun == xml_fun.id:
                    allocated_fun_list.add(xml_fun)
        for func in allocated_fun_list:
            current_fun_list = get_input_or_output_fun_and_fun_elem(func, direction, True, **kwargs)
            for sub in current_fun_list:
                if sub and sub[1] not in [f.name for f in allocated_fun_list]:
                    in_or_out_list.append(sub)
        if unmerged:
            return in_or_out_list
        return merge_list_per_cons_prod(in_or_out_list)
    # For Function()
    if direction == 'output':
        in_or_out_list = get_in_out_function(wanted_object,
                                            kwargs['xml_producer_function_list'],
                                            kwargs['xml_consumer_function_list'])
    else:
        in_or_out_list = get_in_out_function(wanted_object,
                                            kwargs['xml_consumer_function_list'],
                                            kwargs['xml_producer_function_list'])
    if unmerged:
        return in_or_out_list

    return merge_list_per_cons_prod(in_or_out_list)


def merge_list_per_cons_prod(input_list):
    """
    Sorts data's name in alphabetical order and merges list by producer or consumer
    Args:
        input_list ([Data_name, funcion_name]): List of consumer/producer with data per Function

    Returns:
        Sorted + merged list
    """
    input_list = sorted(input_list)
    output_list = []
    empty_dict = {}
    for data, obj_name in input_list:
        if data not in empty_dict:
            output_list.append([data, obj_name])
            empty_dict[data] = len(empty_dict)
        else:
            if obj_name:
                if obj_name not in output_list[empty_dict[data]][1]:
                    output_list[empty_dict[data]][1] += '\\n' + obj_name

    return output_list


def get_in_out_function(wanted_object, wanted_relationship, opposite_wanted_relationship):
    """
    Return the list of Input/Output for Function's objects.
    Args:
        wanted_object (Function): Object's wanted
        wanted_relationship ([Xml_list]): For Output (producer's list) and Input (consumer's list)
        opposite_wanted_relationship ([Xml_list]): slef-explained
    Returns:
        output_list ([Data, prod/cons])
    """
    output_list = []
    flow_list = get_consumes_produces_info(wanted_object, wanted_relationship)
    if not flow_list:
        return output_list
    for data in flow_list:
        if not any(data in s for s in opposite_wanted_relationship):
            output_list.append([data, None])
        else:
            for elem in opposite_wanted_relationship:
                check = True
                if elem[0] == data:
                    for child in elem[1].child_list:
                        if [elem[0], child] in opposite_wanted_relationship:
                            check = False
                    if check:
                        output_list.append([elem[0], elem[1].name])

    return output_list


def get_objects_from_id_list(id_list, object_list):
    """From a list of id (from same type) returns list of corresponding objects"""
    output_list = set()
    if not id_list:
        return output_list

    for wanted_object in object_list:
        if wanted_object.id in id_list:
            output_list.add(wanted_object)

    return output_list


def get_objects_names(xml_object_list):
    """
    Method that returns a list with all object aliases/names from object's list

    """
    object_name_list = []
    # Create the xml [object_name (and object_alias)] list
    for xml_object in xml_object_list:
        object_name_list.append(xml_object.name)
        try:
            if len(xml_object.alias) > 0:
                object_name_list.append(xml_object.alias)
        except AttributeError:
            # To avoid error when there is no alias attribute for the object
            pass

    return object_name_list


def check_parentality(object_a, object_b):
    """Check recursively if object 'a' is not parent of object 'b'"""
    if object_b.parent:
        if object_a == object_b.parent:
            return True
        else:
            return check_parentality(object_a, object_b.parent)
    else:
        return False


def get_children(element, function_list=None, parent_dict=None, count=None, level=None):
    """Get children recursively, adds them to function_list and create parend_dict"""
    if function_list is None:
        function_list = set()
    if parent_dict is None:
        parent_dict = {}
    if not count:
        count = 0

    function_list.add(element)
    if element.child_list:
        count += 1
        if level:
            if (count - 1) == level:
                element.child_list.clear()
                return function_list, parent_dict
        for child in element.child_list:
            parent_dict[child.id] = element.id
            get_children(child, function_list, parent_dict, count, level)

    return function_list, parent_dict


def check_not_family(object_a, object_b):
    """Returns True if object_a and object_b are not in the same family"""
    if not check_parentality(object_a, object_b) and not check_parentality(object_b, object_a):
        return True
    else:
        return False


def get_object_type(object_to_check):
    """From an object, returns its type as string"""
    obj_type_list = [(datamodel.State, 'state'), (datamodel.Function, 'function'),
                     (datamodel.Data, 'data'), (datamodel.FunctionalElement, 'Functional element'),
                     (datamodel.View, 'View'), (datamodel.Transition, 'Transition'),
                     (datamodel.Attribute, 'Attribute'),
                     (datamodel.FunctionalInterface, 'Functional interface'),
                     (datamodel.PhysicalElement, 'Physical element'),
                     (datamodel.PhysicalInterface, 'Physical interface')]
    object_type = ''
    for elem in obj_type_list:
        if isinstance(object_to_check, elem[0]):
            object_type = elem[1]
            return object_type
    return object_type


def get_pandas_table(data_dict):
    """Returns pandas data frame called from command_parser.matched_list()
    with data_dict as list of ..."""
    if 'columns' in data_dict.keys():
        data_frame = pd.DataFrame(data_dict['data'], columns=data_dict['columns'])
    else:
        data_frame = pd.DataFrame(data_dict['data'])
    data_frame = data_frame.T
    if 'Data' in data_dict['title'] or 'Transition' in data_dict['title']:
        if 'Data' in data_dict['title']:
            first = 1
            last = 5
        else:
            first = 3
            last = 4
        for idx in range(first, last):
            data_frame.iloc[idx] = data_frame.iloc[idx].str.join("\\n")
    data_frame = data_frame.style \
        .set_caption(data_dict['title']) \
        .set_properties(**{'white-space': 'nowrap'})

    return data_frame.to_html().replace("\\n", "<br>")



"""Module to test sequece diagrams"""
from conftest import get_jarvis4se, remove_xml_file
import plantuml_adapter

jarvis4se = get_jarvis4se()


def test_fun_inter_simple_sequence(mocker):
    """Notebook equivalent:
    %%jarvis
    with fun_inter_simple_sequence
    Fun_inter is a functional interface
    Fun_elem_1 is a functional element
    Fun_elem_2 is a functional element
    A is a data
    B is a data
    C is a data
    F1 is a function
    F2 is a function
    F1 produces A
    F1 produces C
    F2 consumes C
    F2 produces B
    F1 consumes B
    F2 consumes A
    C implies B
    B implies A
    Fun_elem_1 allocates F1
    Fun_elem_2 allocates F2
    Fun_elem_1 exposes Fun_inter
    Fun_elem_2 exposes Fun_inter
    Fun_inter allocates A
    Fun_inter allocates B
    Fun_inter allocates C

    show sequence Fun_inter
     """
    spy = mocker.spy(plantuml_adapter, "get_sequence_diagram")
    file_name = "fun_inter_simple_sequence"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "Fun_inter is a functional interface\n"
                         "Fun_elem_1 is a functional element\n"
                         "Fun_elem_2 is a functional element\n"
                         "A is a data\n"
                         "B is a data\n"
                         "C is a data\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "F1 produces A\n"
                         "F1 produces C\n"
                         "F2 consumes C\n"
                         "F2 produces B\n"
                         "F1 consumes B\n"
                         "F2 consumes A\n"
                         "C implies B\n"
                         "B implies A\n"
                         "Fun_elem_1 allocates F1\n"
                         "Fun_elem_2 allocates F2\n"
                         "Fun_elem_1 exposes Fun_inter\n"
                         "Fun_elem_2 exposes Fun_inter\n"
                         "Fun_inter allocates A\n"
                         "Fun_inter allocates B\n"
                         "Fun_inter allocates C\n"
                         "\n"
                         "show sequence Fun_inter\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['!pragma teoz true\n',
                'participant fun_elem_1 <<Functional element>>\n',
                'participant fun_elem_2 <<Functional element>>\n',
                'activate fun_elem_1\n',
                'activate fun_elem_2\n',
                'fun_elem_1 -> fun_elem_2 : 1- A\n',
                'fun_elem_2 -> fun_elem_1 : 2- B\n',
                'fun_elem_1 -> fun_elem_2 : 3- C\n',
                'deactivate fun_elem_1\n',
                'deactivate fun_elem_2\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) == len(''.join(expected))



# Libraries
import os


# Modules


class Config:
    """@ingroup tools
    @anchor Config
    Jarvis configurator
    """

    is_log_file = False
    is_diagram_file = False

    @classmethod
    def read(cls):
        """ Read the configuration file if any
        @return None
        """
        try:
            if os.path.isdir("config"):
                with open("config/config.cfg", "r", encoding='utf-8') as file_reader:
                    line = file_reader.readline()

                    error = False
                    while (line != '') & (not error):
                        lines = line.split('=')
                        if len(lines) > 1:
                            if lines[0].strip() == 'log':
                                if 'True' == lines[1].strip() or '1' == lines[1].strip():
                                    cls.is_log_file = True
                                    # Logger depends from Config. Thus simple print
                                    print("Log file storage activated")
                            elif lines[0].strip() == 'diagram':
                                if 'True' == lines[1].strip() or '1' == lines[1].strip():
                                    cls.is_diagram_file = True
                                    # Logger depends from Config. Thus simple print
                                    print("Diagram file storage activated")
                            else:
                                error = True
                        else:
                            error = True
                        line = file_reader.readline()

                    if error:
                        # Logger depends from Config. Thus simple print
                        print("[ERROR] Configuration file is not correctly formatted")
            else:
                # Logger depends from Config. Thus simple print
                print(f"No configuration file detected in {os.getcwd()}\\config")
        except EnvironmentError as ex:
            # Logger depends from Config. Thus simple print
            print(f"[ERROR] Unable to read the configuration file in {os.getcwd()}\\config: {ex}")



"""Module to tests decomposition diagrams"""
import io

from conftest import get_jarvis4se, remove_xml_file
import plantuml_adapter

jarvis4se = get_jarvis4se()


def test_function_with_childs_decomposition(mocker, function_with_childs_cell):
    """See Issue #5, Notebook equivalent:
    %%jarvis
    with function_with_childs_decomposition
    F1 is a function
    F1a is a function
    F1b is a function
    F1c is a function
    F1d is a function
    F1e is a function
    F2 is a function
    F3 is a function

    F1 is composed of F1a
    F1 is composed of F1b
    F1 is composed of F1c
    F1 is composed of F1d
    F1 is composed of F1e

    a is a data
    F1 produces a
    F2 consumes a

    F1a produces a
    F1b consumes a

    b is a data
    F1c produces b
    F1d consumes b

    c is a data
    F3 produces c
    F1e consumes c

    show decomposition F1
     """
    spy = mocker.spy(plantuml_adapter, "get_function_diagrams")
    file_name = "function_with_childs_decomposition"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         f"{function_with_childs_cell}\n"
                         "show decomposition F1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['object "F3" as f3 <<Function>>\n',
                'component "F1" as f1 <<Function>>{\n',
                'object "F1c" as f1c <<Function>>\n',
                'object "F1d" as f1d <<Function>>\n',
                'object "F1e" as f1e <<Function>>\n',
                'object "F1a" as f1a <<Function>>\n',
                'object "F1b" as f1b <<Function>>\n',
                '}\n',
                'object "F2" as f2 <<Function>>\n',
                'f1a #--> f2 : a\n',
                'f1c #--> f1d : b\n',
                'f1a #--> f1b : a\n',
                'f3 #--> f1e : c\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 8*len("\'id: xxxxxxxxxx\n")


def test_fun_elem_decompo_with_no_flow(mocker, monkeypatch):
    """Notebook equivalent (see Issue_#7):
    %%jarvis
    with fun_elem_decompo_with_no_flow
    F1 is a function
    F1a is a function
    F1b is a function
    F1c is a function
    F1 is composed of F1a
    F1 is composed of F1b
    F1 is composed of F1c

    F1c1 is a function
    F1c is composed of F1c1

    F2 is a function
    F2a is a function
    F2 is composed of F2a

    F3 is a function
    F3a is a function
    F3 is composed of F3a

    E1 is a functional element
    E1a is a functional element
    E1b is a functional element
    E1c is a functional element
    E1 is composed of E1a
    E1 is composed of E1b
    E1 is composed of E1c
    E1c1 is a functional element
    E1c2 is a functional element
    E1c is composed of E1c1
    E1c is composed of E1c2

    E1 allocates F1
    E1 allocates F2
    E1a allocates F1a
    E1a allocates F3a
    E1b allocates F1b
    E1c allocates F1c
    E1c1 allocates F1c1

    show decomposition E1
     """
    monkeypatch.setattr('sys.stdin', io.StringIO('y'))  # Say yes for adding F3 alloacted to E1
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_decomposition")
    file_name = "fun_elem_decompo_with_no_flow"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "F1a is a function\n"
                         "F1b is a function\n"
                         "F1c is a function\n"
                         "F1 is composed of F1a\n"
                         "F1 is composed of F1b\n"
                         "F1 is composed of F1c\n"
                         "\n"
                         "F1c1 is a function\n"
                         "F1c is composed of F1c1\n"
                         "\n"
                         "F2 is a function\n"
                         "F2a is a function\n"
                         "F2 is composed of F2a\n"
                         "\n"
                         "F3 is a function\n"
                         "F3a is a function\n"
                         "F3 is composed of F3a\n"
                         "\n"
                         "E1 is a functional element\n"
                         "E1a is a functional element\n"
                         "E1b is a functional element\n"
                         "E1c is a functional element\n"
                         "E1 is composed of E1a\n"
                         "E1 is composed of E1b\n"
                         "E1 is composed of E1c\n"
                         "E1c1 is a functional element\n"
                         "E1c2 is a functional element\n"
                         "E1c is composed of E1c1\n"
                         "E1c is composed of E1c2\n"
                         "\n"
                         "E1 allocates F1\n"
                         "E1 allocates F2\n"
                         "E1a allocates F1a\n"
                         "E1a allocates F3a\n"
                         "E1b allocates F1b\n"
                         "E1c allocates F1c\n"
                         "E1c1 allocates F1c1\n"
                         "\n"
                         "show decomposition E1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E1" as e1 <<Functional element>>{\n',
                'object "F2" as f2 <<Function>>\n',
                'component "E1b" as e1b <<Functional element>>{\n',
                'object "F1b" as f1b <<Function>>\n',
                '}\n',
                'component "E1a" as e1a <<Functional element>>{\n',
                'object "F1a" as f1a <<Function>>\n',
                'object "F3a" as f3a <<Function>>\n',
                '}\n',
                'component "E1c" as e1c <<Functional element>>{\n',
                'component "E1c1" as e1c1 <<Functional element>>{\n',
                'object "F1c1" as f1c1 <<Function>>\n',
                '}\n',
                'component "E1c2" as e1c2 <<Functional element>>{\n',
                '}\n',
                '}\n',
                '}\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 11*len("\'id: xxxxxxxxxx\n")


def test_fun_elem_decompo_with_no_childs(mocker):
    """See Issue #13, Notebook equivalent:
    %%jarvis
    with fun_elem_decompo_with_no_childs
    F1 is a function
    F2 is a function
    F1 is composed of F2
    E1 is a functional element
    E1 allocates F1

    show decomposition E1
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_decomposition")
    file_name = "fun_elem_decompo_with_no_childs"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "F1 is composed of F2\n"
                         "E1 is a functional element\n"
                         "E1 allocates F1\n"
                         "\n"
                         "show decomposition E1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E1" as e1 <<Functional element>>{\n',
                '}\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == len("\'id: xxxxxxxxxx\n")


def test_fun_elem_decompo_with_interface(mocker):
    """Notebook equivalent:
    %%jarvis
    with fun_elem_decompo_with_interface
    F1 is a function
    F1b is a function
    F1c is a function
    F1 is composed of F1b
    F1 is composed of F1c

    F1c1 is a function
    F1c is composed of F1c1

    F_ext is a function

    E1 is a functional element
    E1b is a functional element
    E1c is a functional element
    Ext is a functional element
    E1 is composed of E1c
    E1 is composed of E1b
    E1c1 is a functional element
    E1c is composed of E1c1

    E1 allocates F1
    E1b allocates F1b
    E1c allocates F1c
    E1c1 allocates F1c1
    Ext allocates F_ext

    A is a data
    A_2 is a data
    B is a data
    C is a data

    Fun_inter_A is a functional interface
    Fun_inter_A allocates A
    Fun_inter_A allocates A_2
    Fun_inter_B is a functional interface
    Fun_inter_B allocates B

    F1c1 produces A
    F1c1 produces A_2
    F1b consumes A
    F1b consumes A_2

    F_ext produces B
    F1c1 consumes B

    F1b produces C
    F_ext consumes C

    E1c exposes Fun_inter_A
    E1c1 exposes Fun_inter_A
    E1b exposes Fun_inter_A

    show decomposition E1
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_decomposition")
    file_name = "fun_elem_decompo_with_interface"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "F1b is a function\n"
                         "F1c is a function\n"
                         "F1 is composed of F1b\n"
                         "F1 is composed of F1c\n"
                         "\n"
                         "F1c1 is a function\n"
                         "F1c is composed of F1c1\n"
                         "\n"
                         "F_ext is a function\n"
                         "\n"
                         "E1 is a functional element\n"
                         "E1b is a functional element\n"
                         "E1c is a functional element\n"
                         "Ext is a functional element\n"
                         "E1 is composed of E1b\n"
                         "E1 is composed of E1c\n"
                         "E1c1 is a functional element\n"
                         "E1c is composed of E1c1\n"
                         "\n"
                         "E1 allocates F1\n"
                         "E1b allocates F1b\n"
                         "E1c allocates F1c\n"
                         "E1c1 allocates F1c1\n"
                         "Ext allocates F_ext\n"
                         "\n"
                         "A is a data\n"
                         "A_2 is a data\n"
                         "B is a data\n"
                         "C is a data\n"
                         "\n"
                         "Fun_inter_A is a functional interface\n"
                         "Fun_inter_A allocates A\n"
                         "Fun_inter_A allocates A_2\n"
                         "Fun_inter_B is a functional interface\n"
                         "Fun_inter_B allocates B\n"
                         "\n"
                         "F1c1 produces A\n"
                         "F1c1 produces A_2\n"
                         "F1b consumes A\n"
                         "F1b consumes A_2\n"
                         "\n"
                         "F_ext produces B\n"
                         "F1c1 consumes B\n"
                         "\n"
                         "F1b produces C\n"
                         "F_ext consumes C\n"
                         "\n"
                         "E1c exposes Fun_inter_A\n"
                         "E1c1 exposes Fun_inter_A\n"
                         "E1b exposes Fun_inter_A\n"
                         "\n"
                         "show decomposition E1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E1" as e1 <<Functional element>>{\n',
                'component "E1c" as e1c <<Functional element>>{\n',
                'component "E1c1" as e1c1 <<Functional element>>{\n',
                'object "F1c1" as f1c1 <<Function>>\n',
                '}\n',
                '}\n',
                'component "E1b" as e1b <<Functional element>>{\n',
                'object "F1b" as f1b <<Function>>\n',
                '}\n',
                '}\n',
                'component "Ext" as ext <<Functional element>>{\n',
                'object "F_ext" as f_ext <<Function>>\n',
                '}\n',
                'f_ext #--> f1c1 : B\n',
                'f1b #--> f_ext : C\n',
                'e1b', ' -- ', 'e1c1 ', ': fun_inter_a\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 8*len("\'id: xxxxxxxxxx\n")



"""Test module with tests associated to question_answer.py"""
from conftest import get_jarvis4se, remove_xml_file
from jarvis import question_answer

jarvis4se = get_jarvis4se()


def test_list_data_functional_interface(mocker):
    """See issue #55 Notebook equivalent:
    %%jarvis
    with list_data_functional_interface
    F1 is a function
    F2 is a function
    F3 is a function

    a is a data
    F1 produces a
    F2 consumes a
    F3 consumes a

    E1 is a functional element
    E1 allocates F1

    E2 is a functional element
    E2 allocates F2

    E3 is a functional element
    E3 allocates F3

    I_E1_E2 is a functional interface
    I_E1_E2 allocates a
    E1 exposes I_E1_E2
    E2 exposes I_E1_E2

    list data I_E1_E2
     """
    spy = mocker.spy(question_answer, "switch_data")
    file_name = "list_data_functional_interface"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "F3 is a function\n"
                         "\n"
                         "a is a data\n"
                         "F1 produces a\n"
                         "F2 consumes a\n"
                         "F3 consumes a\n"
                         "\n"
                         "E1 is a functional element\n"
                         "E1 allocates F1\n"
                         "\n"
                         "E2 is a functional element\n"
                         "E2 allocates F2\n"
                         "\n"
                         "E3 is a functional element\n"
                         "E3 allocates F3\n"
                         "\n"
                         "I_E1_E2 is a functional interface\n"
                         "I_E1_E2 allocates a\n"
                         "E1 exposes I_E1_E2\n"
                         "E2 exposes I_E1_E2\n"
                         "\n"
                         "list data I_E1_E2\n")

    result = spy.spy_return['data']
    expected = [{'Data': 'a',
                 'Last consumer Function(s)': ['F2'],
                 'Last consumer Functional element(s)': ['E2'],
                 'Last producer Function(s)': ['F1'],
                 'Last producer Functional element(s)': ['E1']}]

    remove_xml_file(file_name)

    assert expected == result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Notebook interface, this module contains ipython Magic class with cell_magic as entry point of
jarvi4se"""
# Libraries
import re
import os
import getpass
import shutil
from datetime import datetime
from io import StringIO

from IPython.core.magic import (Magics, magics_class, cell_magic)

# Modules
from xml_adapter import GenerateXML, XmlParser3SE
from tools import Logger


# The class MUST call this class decorator at creation time
@magics_class
class MagicJarvis(Magics):
    """Magic iPython class"""
    def __init__(self, shell, parser):
        # You must call the parent constructor
        super().__init__(shell)
        self.parser = parser

    @cell_magic
    def jarvis(self, _, cell):
        """Entry point for jarvi4se"""
        # Initialize xml_parser i.e. empty obj_dict that will contain objet's lists
        xml_parser = XmlParser3SE()
        obj_dict = xml_parser.xml_dict
        # We create a string buffer containing the
        # contents of the cell.
        sio = StringIO(cell)
        # Take the value within the buffer
        input_str = sio.getvalue()
        # Delete the '"' from input string, to avoid xml to plantuml errors.
        input_str = input_str.replace('"', "")
        # Delete extra whitespaces
        input_str = re.sub(' +', ' ', input_str)
        # Get model's declaration, need a space after "with" otherwise print a message
        xml_name_str = re.match(r"^with (.*)(?=.|\n)", input_str, re.MULTILINE)
        if xml_name_str:
            xml_name = xml_name_str.group(1)
            # If the model(i.e. file) already exists, parse it to extract lists
            if os.path.isfile(f"{xml_name}.xml"):
                obj_dict = xml_parser.parse_xml(f"{xml_name}.xml")

                # TODO : why this print ?
                if isinstance(obj_dict, str):
                    print(obj_dict)
                    return

                Logger.set_info(__name__,
                                f"{xml_name}.xml parsed")
                output_xml = GenerateXML(f"{xml_name}.xml")
            # Else create an empty xml_lists
            # or will be named by default "Outpout"
            else:
                if len(xml_name) > 1:
                    Logger.set_info(__name__,
                                    f"Creating {xml_name}.xml !")

                    output_xml = GenerateXML(f"{xml_name}.xml")
                else:
                    Logger.set_info(__name__,
                                    "Xml's file does not exists, creating it ('output.xml' by default) !")
                    output_xml = GenerateXML("")
                output_xml.write()

            obj_dict['output_xml'] = output_xml
            update = self.parser.lookup_table(input_str, **obj_dict)

            if not update:
                return

            if 1 in update:
                Logger.set_info(__name__,
                                f"{output_xml.file} updated")
            else:
                Logger.set_info(__name__,
                                f"No update for {output_xml.file}")
        else:
            Logger.set_error(__name__,
                             "Bad model's declaration, model's name should be written or add a ' '(blank space) "
                             "after 'with' command to create default 'Output.xml'")


def greet_user():
    """Greets the user according to the time thanks to :
    https://ireadblog.com/posts/141/how-to-build-your-personal-ai-assistant-using-python"""
    hour = datetime.now().hour
    # Use getpass() because available on Unix/Windows
    user_name = getpass.getuser()

    # Single display (not related to logging)
    if 6 <= hour < 12:
        print(f"Good morning {user_name}")
    elif 12 <= hour < 16:
        print(f"Good afternoon {user_name}")
    elif 16 <= hour < 19:
        print(f"Good evening {user_name}")
    else:
        print(f"Hello {user_name}")

    print("I am Jarvis. How may I assist you?")


def clean_folders():
    """ Clean Jarvis folder
    @return None
    """
    for folder in ["./diagrams", "./log"]:
        if os.path.isdir(folder):
            for filename in os.listdir(folder):
                file_path = os.path.join(folder, filename)
                try:
                    if os.path.isfile(file_path) or os.path.islink(file_path):
                        os.remove(file_path)
                    elif os.path.isdir(file_path):
                        shutil.rmtree(file_path)
                except Exception as ex:
                    # function called before Logger initialization.
                    print('[ERROR] Failed to delete %s. Reason: %s' % (file_path, ex))



# Libraries

# Modules


def get_hyperlink(path):
    """@anchor get_hyperlink
    Convert file path into HTML link
    @param[in] path file path
    @return HTML link
    """
    text = "Click to open in new tab"
    # convert the url into link
    return f'<a href="{path}" target="_blank">{text}</a>'



# Libraries
from sys import version as py_ver
from importlib.metadata import version
from IPython.core.magic import (Magics, magics_class, cell_magic, line_magic)
from IPython.display import display, HTML, Markdown

# Modules
from .util import get_hyperlink


@magics_class
class MagicTools(Magics):
    """@ingroup tools
    @anchor MagicTools
    Magic call for IPython interface
    """

    def __init__(self, shell, generator):
        """
        @var generator
        Jarvis diagram generator
        """

        # You must call the parent constructor
        super().__init__(shell)
        self.generator = generator

    @staticmethod
    @line_magic
    def retrieve_pkg_version(_):
        """Magic line that get Jarvis dependencies versions
        @return None
        """
        pkg = ['ipython', 'lxml', 'notebook', 'plantuml', 'jarvis4se', 'pandas', 'requests']
        pkg_ver =  "\n".join(['=='.join(tups) for tups in zip(pkg, map(version, pkg))])
        print(pkg_ver, "\npython=={}".format(py_ver[:6]))

    @cell_magic
    def diagram(self, _, cell):
        """Magic cell that allows to call directly Jarvis diagram generator
        @param[in] cell cell instance
        @param[in] _ not used
        @return None
        """
        out = self.generator.get_diagram_url(cell, True)
        if out:
            hyper = get_hyperlink(out)
            display(HTML(hyper))
            # Single display (not related to logging)
            print("Overview :")
            display(Markdown(f'![figure]({out})'))



"""Test module for tools pkg i.e. MagicTools"""
from IPython import get_ipython

import jarvis

ip = get_ipython()
generator = jarvis.PlantUmlGen()
my_magic = jarvis.MagicTools(ip, generator)


def test_retrieve_pkg_version(capsys):
    """Notebook equivalent:
    %retrieve_pkg_version
    """
    my_magic.retrieve_pkg_version('')
    captured = capsys.readouterr()
    expected = ["lxml==", "notebook==", "plantuml==", "jarvis4se==", "pandas==", "python=="]
    assert all(i in captured.out for i in expected)


def test_diagram_cell(capsys, mocker):
    """Notebook equivalent:
    %%diagram
    @startuml
    !define Junction_Or circle #black
    !define Junction_And circle #whitesmoke

    Junction_And JunctionAnd
    Junction_Or JunctionOr

    archimate #Technology "VPN Server" as vpnServerA <<technology-device>>

    rectangle GO #lightgreen
    rectangle STOP #red
    rectangle WAIT #orange
    GO -up-> JunctionOr
    STOP -up-> JunctionOr
    STOP -down-> JunctionAnd
    WAIT -down-> JunctionAnd
    @enduml
    """
    spy = mocker.spy(generator, "get_diagram_url")
    my_magic.diagram('', "@startuml\n"
                         "!define Junction_Or circle #black\n"
                         "!define Junction_And circle #whitesmoke\n"
                         "\n"
                         "\n"
                         "Junction_And JunctionAnd\n"
                         "Junction_Or JunctionOr\n"
                         "\n"
                         "archimate #Technology 'VPN Server' as vpnServerA <<technology-device>>\n"
                         "\n"
                         "rectangle GO #lightgreen\n"
                         "rectangle STOP #red\n"
                         "rectangle WAIT #orange\n"
                         "GO -up-> JunctionOr\n"
                         "STOP -up-> JunctionOr\n"
                         "STOP -down-> JunctionAnd\n"
                         "WAIT -down-> JunctionAnd\n"
                         "@enduml\n")
    expected_plantuml_link = "http://www.plantuml.com/plantuml/svg/TL2nhi8m3Dpz5KOTcFe7gA8JUWmK" \
                             "YGf651BJHgGESjCY_fx04oW8sExEToVRypue2KFdO6BeQ9bmER0ErlE-4jHMj2FC3ax" \
                             "fqwUZPFEoN5eRgE_yYG3WpV4a4KDQ_iIL02ZHhUrKY4KrwPQzyyqLfzlr2ZSa8yaKLO" \
                             "_ZcVzPYRDPUFboGwFLL1G0GZeeRk92YmepPvisD4B4oM1JLslCX4oYxSg_6ZClaH74P" \
                             "3wSyo9Ty17weHf_uKI_d_de-pQO4vlxisy="
    expected_notebook_output = "<IPython.core.display.HTML object>\n" \
                               "Overview :\n" \
                               "<IPython.core.display.Markdown object>\n"
    captured = capsys.readouterr()
    assert spy.spy_return == expected_plantuml_link
    assert captured.out == expected_notebook_output



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Module with methods relative to Functional section"""
# Libraries
import re

# Modules
import datamodel
from . import shared_orchestrator
from .question_answer import get_objects_names, get_children, check_get_object, check_parentality
from tools import Logger


def check_add_predecessor(data_predecessor_str_set, xml_data_list, xml_view_list, output_xml):
    """
    Check if each string in data_predecessor_str_set is corresponding to an actual Data object,
    create new [Data, predecessor] objects lists for object's type : Data.
    Send lists to add_predecessor() to write them within xml and then returns update_list from it.

        Parameters:
            data_predecessor_str_set ([str]) : Lists of string from jarvis cell
            xml_data_list ([Data]) : Data list from xml parsing
            xml_view_list ([View]) : View list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update ([0/1]) : 1 if update, else 0
    """
    data_predecessor_list = []

    allocated_item_list = []
    # Filter input string
    data_predecessor_str_list = shared_orchestrator.cut_string_list(data_predecessor_str_set)

    # Create data names list already in xml
    xml_data_name_list = get_objects_names(xml_data_list)

    for elem in data_predecessor_str_list:
        is_elem_found = True
        if elem[0] not in xml_data_name_list:
            is_elem_found = False
            if elem[1] not in xml_data_name_list:
                Logger.set_error(__name__,
                                f"{elem[0]} and {elem[1]} do not exist")
            else:
                Logger.set_error(__name__,
                                f"{elem[0]} does not exist")

        if elem[0] in xml_data_name_list:
            if elem[1] not in xml_data_name_list:
                is_elem_found = False
                Logger.set_error(__name__,
                                f"{elem[1]} does not exist")

        if is_elem_found:
            predecessor = None
            selected_data = None
            existing_predecessor_id_list = []
            for data in xml_data_list:
                if elem[0] == data.name:
                    selected_data = data
                    for existing_predecessor in data.predecessor_list:
                        existing_predecessor_id_list.append(existing_predecessor.id)
            for da in xml_data_list:
                if elem[1] == da.name and da.id not in existing_predecessor_id_list:
                    predecessor = da
            if predecessor is not None and selected_data is not None:
                data_predecessor_list.append([selected_data, predecessor])
            allocation_chain_1 = shared_orchestrator.check_add_allocated_item(elem[0],
                                                                              xml_data_list,
                                                                              xml_view_list)
            if allocation_chain_1:
                allocated_item_list.append(allocation_chain_1)
            allocation_chain_2 = shared_orchestrator.check_add_allocated_item(elem[1],
                                                                              xml_data_list,
                                                                              xml_view_list)
            if allocation_chain_2:
                allocated_item_list.append(allocation_chain_2)

    update = add_predecessor(data_predecessor_list, xml_data_list, output_xml)
    shared_orchestrator.add_allocation({5: allocated_item_list}, output_xml)

    return update


def add_predecessor(predecessor_list, xml_data_list, output_xml):
    """
    Check if input lists is not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            predecessor_list ([Data, Data(predecessor)]) : Data object to set new predessor and
            predecessor Data
            xml_data_list ([Data]) : Data list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """

    if not predecessor_list:
        return 0

    output_xml.write_predecessor(predecessor_list)
    # Warn the user once added within xml
    for data_predecessor in predecessor_list:
        for d in xml_data_list:
            if data_predecessor[0].id == d.id:
                d.add_predecessor(data_predecessor[1])
                Logger.set_info(__name__,
                               f"{data_predecessor[1].name} predecessor for "
                               f"{data_predecessor[0].name}")
    return 1


def check_add_consumer_function(consumer_str_list, xml_consumer_function_list,
                                xml_producer_function_list, xml_function_list, xml_data_list,
                                output_xml):
    """
    Check if each string in consumer_str_list are corresponding to an actual object, create new
    [data, consumer] objects list for object's type : Function.
    Send lists to add_consumer_function() to write them within xml and then returns update_list
    from it.

        Parameters:
            consumer_str_list ([str]) : Lists of string from jarvis cell
            xml_consumer_function_list ([Data_name_str, Function]) : Data's name and consumer's
            function list from xml
            xml_producer_function_list ([Data_name_str, Function]) : Data's name and producer's
            function list from xml
            xml_function_list ([Function]) : Function list from xml parsing
            xml_data_list ([Data]) : Data list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    new_consumer_list = []
    # Create object names/aliases list and data's name
    xml_function_name_list = get_objects_names(xml_function_list)
    xml_data_name_list = get_objects_names(xml_data_list)
    # Loop to filter consumer and create a new list
    for elem in consumer_str_list:
        is_elem_found = True
        if not any(item == elem[1] for item in xml_function_name_list) and \
                not any(item == elem[0] for item in xml_data_name_list):
            is_elem_found = False
            Logger.set_error(__name__,
                            f"{elem[1]} and {elem[0]} do not exist")
        elif not any(item == elem[1] for item in xml_function_name_list) or \
                not any(item == elem[0] for item in xml_data_name_list):
            is_elem_found = False
            if any(item == elem[1] for item in xml_function_name_list) and \
                    not any(item == elem[0] for item in xml_data_name_list):
                Logger.set_error(__name__,
                                f"{elem[0]} does not exist")
            elif any(item == elem[0] for item in xml_data_name_list) and \
                    not any(item == elem[1] for item in xml_function_name_list):
                Logger.set_error(__name__,
                                f"{elem[1]} does not exist")

        if is_elem_found:
            # Loop to filter consumer and create a new list
            for function in xml_function_list:
                if elem[1] == function.name or elem[1] == function.alias:
                    if [elem[0], function] not in xml_consumer_function_list:
                        if [elem[0], function] not in xml_producer_function_list:
                            new_consumer_list.append([elem[0], function])
                            parent = add_parent_recursively(elem[0], function,
                                                            xml_consumer_function_list,
                                                            xml_producer_function_list,
                                                            new_consumer_list,
                                                            output_xml, "consumer")

                            if parent is not None:
                                for par in parent:
                                    if par:
                                        new_consumer_list.append(par)
                        elif [elem[0], function] in xml_producer_function_list:
                            pass

    update = add_consumer_function(new_consumer_list, xml_consumer_function_list, output_xml)

    return update


def add_consumer_function(new_consumer_list, xml_consumer_function_list, output_xml):
    """
    Check if input list is not empty, write in xml for each element and return update list if some
    updates has been made

        Parameters:
            new_consumer_list ([Data_name_str, Function]) : Data's name and consumer's function list
            xml_consumer_function_list ([Data_name_str, Function]) : Data's name and consumer's
            function list from xml
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """

    if not new_consumer_list:
        return 0

    output_xml.write_consumer(new_consumer_list)
    # Warn the user once added within xml
    for consumer in new_consumer_list:
        xml_consumer_function_list.append(consumer)
        Logger.set_info(__name__,
                       f"{consumer[1].name} consumes {consumer[0]}")

    return 1


def add_parent_recursively(flow, function, current_list, opposite_list, new_list, output_xml,
                           relationship_str, out=False):
    """
    Recursive method around add_parent_for_data().
        Parameters:
            flow (Data_name_str) : Data's name
            function (Function) : Current function's parent
            current_list ([Data_name_str, function_name_str]) : 'Current' list (producer/consumer)
            opposite_list ([Data_name_str, function_name_str]) : Opposite list from current
            new_list ([Data_name_str, Function]) : Data's name and consumer/producer's function list
            output_xml (GenerateXML object) : XML's file object
            relationship_str (str) : "consumer" or "producer"
            out (bool) : List for recursivity
        Returns:
            elem ([data, Function]) : Return parent

    """
    if not out:
        out = []
    parent = add_parent_for_data(flow, function,
                                 current_list,
                                 opposite_list,
                                 new_list,
                                 output_xml, relationship_str)

    if parent is not None:
        out.append(parent)
        return add_parent_recursively(flow, function.parent, current_list, opposite_list, new_list,
                                      output_xml,
                                      relationship_str, out)

    return out


def add_parent_for_data(flow, function, current_list, opposite_list, new_list, output_xml,
                        relationship_str):
    """
    Adds direct parent's function of consumer/producer and delete internal flow if the case

        Parameters:
            flow (Data_name_str) : Data's name
            function (Function) : Current function's parent
            current_list ([Data_name_str, function_name_str]) : 'Current' list (producer/consumer)
            opposite_list ([Data_name_str, function_name_str]) : Opposite list from current
            new_list ([Data_name_str, Function]) : Data's name and consumer/producer's function list
            output_xml (GenerateXML object) : XML's file object
            relationship_str (str) : "consumer" or "producer"
        Returns:
            elem ([data, Function]) : Return parent
    """
    elem = [flow, function.parent]
    temp_set = set()
    check = False

    if function.parent is not None and elem not in [*current_list, *new_list]:
        parent_child_list, parent_child_dict = get_children(function.parent)

        current_loop_check = False
        for current_loop_data in [*current_list, *new_list]:
            if current_loop_data[0] == flow and current_loop_data[1] not in parent_child_list:
                current_loop_check = True
        for data_function in opposite_list:
            if data_function[0] == flow:
                temp_set.add(data_function[1])
        length = len(temp_set)
        if temp_set == set():
            check = True
        else:
            for fun in temp_set:
                if fun not in parent_child_list:
                    check = True
                    if any(s == [flow, function.parent] for s in opposite_list):
                        delete_opposite(flow, function.parent, output_xml, relationship_str)
                if fun in parent_child_list:
                    length -= 1
        if length == 0 and not current_loop_check and temp_set != set():
            delete_opposite(flow, function.parent, output_xml, relationship_str)
    if check:
        return elem


def delete_opposite(data, function, output_xml, relationship_type):
    """
    Delete specific consumer/producer relationship within xml's file.

        Parameters:
            data (Data_name_str) : Data's name
            function (Function) : Current Function object
            output_xml (GenerateXML object) : XML's file object
            relationship_type (str) : Type of relationship (i.e. consumer or producer)
        Returns:
            None
    """

    if relationship_type == "producer":
        output_xml.delete_single_consumer_producer(data,
                                                   function,
                                                   "consumer")
        Logger.set_info(__name__,
                       f"{function.name} does not consume {data} anymore")
    elif relationship_type == "consumer":

        output_xml.delete_single_consumer_producer(data,
                                                   function,
                                                   "producer")
        Logger.set_info(__name__,
                       f"{function.name} does not produce {data} anymore")


def check_add_producer_function(producer_str_list, xml_consumer_function_list,
                                xml_producer_function_list, xml_function_list, xml_data_list,
                                output_xml):
    """
    Check if each string in consumer_str_list are corresponding to an actual object, create new
    [data, producer] objects list for object's type : Function.
    Send list to add_producer_function() to write them within xml and then returns update.

        Parameters:
            producer_str_list ([str]) : List of string from jarvis cell
            xml_consumer_function_list ([Data_name_str, Function]) : Data's name and consumer's
            function list from xml
            xml_producer_function_list ([Data_name_str, Function]) : Data's name and producer's
            function list from xml
            xml_function_list ([Function]) : Function list from xml parsing
            xml_data_list ([Data]) : Data list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    new_producer_list = []
    # Create object names/aliases list
    xml_function_name_list = get_objects_names(xml_function_list)
    xml_data_name_list = get_objects_names(xml_data_list)
    # Loop to filter producer and create a new list
    for elem in producer_str_list:
        is_elem_found = True
        if not any(item == elem[1] for item in xml_function_name_list) and \
                not any(item == elem[0] for item in xml_data_name_list):
            is_elem_found = False
            Logger.set_error(__name__,
                            f"{elem[1]} and {elem[0]} do not exist")
        elif not any(item == elem[1] for item in xml_function_name_list) or \
                not any(item == elem[0] for item in xml_data_name_list):
            is_elem_found = False
            if any(item == elem[1] for item in xml_function_name_list) and \
                    not any(item == elem[0] for item in xml_data_name_list):
                Logger.set_error(__name__,
                                f"{elem[0]} does not exist")
            elif any(item == elem[0] for item in xml_data_name_list) and \
                    not any(item == elem[1] for item in xml_function_name_list):
                Logger.set_error(__name__,
                                f"{elem[1]} does not exist")

        if is_elem_found:
            # Loop to filter consumer and create a new list
            for function in xml_function_list:
                if elem[1] == function.name or elem[1] == function.alias:
                    if [elem[0], function] not in xml_producer_function_list:
                        if [elem[0], function] not in xml_consumer_function_list:
                            new_producer_list.append([elem[0], function])
                            parent = add_parent_recursively(elem[0], function,
                                                            xml_producer_function_list,
                                                            xml_consumer_function_list,
                                                            new_producer_list, output_xml,
                                                            "producer")
                            if parent is not None:
                                for par in parent:
                                    if par:
                                        new_producer_list.append(par)

                        elif [elem[0], function] in xml_consumer_function_list:
                            pass

    update = add_producer_function(new_producer_list, xml_producer_function_list, output_xml)

    return update


def add_producer_function(new_producer_list, xml_producer_function_list, output_xml):
    """
    Check if input list is not empty, write in xml for each element and return update list if some
    updates has been made

        Parameters:
            new_producer_list ([Data_name_str, Function]) : Data's name and producer's function list
            xml_producer_function_list ([Data_name_str, Function]) : Data's name and producer's
            function list from xml
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    if not new_producer_list:
        return 0

    output_xml.write_producer(new_producer_list)
    # Warn the user once added within xml
    for producer in new_producer_list:
        xml_producer_function_list.append(producer)
        Logger.set_info(__name__,
                       f"{producer[1].name} produces {producer[0]}")
    return 1


# TODO: Check condition_str on data and (add LogicalType, ArithmeticType in datamodel.py)
def check_add_transition_condition(trans_condition_str_list, xml_transition_list, output_xml):
    """
    Check if each string in trans_condition_str_list is corresponding to an actual Transition
    object, create new [Transition, condition_str] objects lists for object's type : Transition.
    Send lists to add_transition_condition() to write them within xml and then returns update_list
    from it.

        Parameters:
            trans_condition_str_list ([str]) : Lists of string from jarvis cell
            xml_transition_list ([Transition]) : Transition list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    condition_list = []
    # Create a list with all transition names/aliases already in the xml
    xml_transition_name_list = get_objects_names(xml_transition_list)
    for transition_str, condition_str in trans_condition_str_list:
        is_elem_found = True
        if not any(transition_str in s for s in xml_transition_name_list):
            is_elem_found = False
            Logger.set_error(__name__,
                            f"The transition {transition_str} does not exist")

        if is_elem_found:
            for transition in xml_transition_list:
                if transition_str == transition.name or transition_str == transition.alias:
                    if not condition_str.lstrip(' ') in transition.condition_list:
                        condition_list.append([transition, condition_str.lstrip(' ')])

    update = add_transition_condition(condition_list, output_xml)

    return update


def add_transition_condition(condition_list, output_xml):
    """
    Check if input list is not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            condition_list ([Transition, condition_str]) : Transition object and conditions as str
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    if not condition_list:
        return 0

    output_xml.write_transition_condition(condition_list)
    for elem in condition_list:
        elem[0].add_condition(elem[1])
        Logger.set_info(__name__,
                       f"Condition for {elem[0].name} : {elem[1]}")
    return 1


def check_add_src_dest(src_dest_str, xml_transition_list, xml_state_list, output_xml):
    """
    Check if each string in src_dest_str is corresponding to an actual Transition and State object,
    create new [Transition, State] objects lists.
    Send lists to add_src_dest() to write them within xml and then returns update_list from it.

        Parameters:
            src_dest_str ([str]) : Lists of string from jarvis cell
            xml_transition_list ([Transition]) : Transition list from xml parsing
            xml_state_list ([State]) : State list from xml parsing
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    new_src_list = []
    new_dest_list = []
    # Create lists with all object names/aliases already in the xml
    xml_transition_name_list = get_objects_names(xml_transition_list)
    xml_state_name_list = get_objects_names(xml_state_list)

    concatenated_lists = [*xml_transition_name_list, *xml_state_name_list]

    # elem = [source/destination, transition_name, state_name]
    for elem in src_dest_str:
        is_elem_found = True
        if not all(t in concatenated_lists for t in [elem[1], elem[2]]):
            is_elem_found = False
            if any(elem[1] in s for s in xml_transition_name_list) and not any(
                    elem[2] in j for j in xml_state_name_list):
                Logger.set_error(__name__,
                                f"{elem[2]} state does not exist")
            elif any(elem[2] in s for s in xml_state_name_list) and not any(
                    elem[1] in j for j in xml_transition_name_list):
                Logger.set_error(__name__,
                                f"{elem[1]} transition does not exist")
            else:
                Logger.set_error(__name__,
                                f"{elem[1]} transition and {elem[2]} state do not exist")

        if is_elem_found:
            if elem[0] == "source":
                for transition in xml_transition_list:
                    if elem[1] == transition.name or elem[1] == transition.alias:
                        for state in xml_state_list:
                            if elem[2] == state.name or elem[2] == state.alias:
                                if not isinstance(state.type, datamodel.BaseType):
                                    if 'EXIT' in state.type.name: 
                                        Logger.set_error(__name__,
                                                        f"{elem[2]} is typed as EXIT state, "
                                                        f"it cannot be put as source's transition (not added)")
                                else:
                                    if transition.source != state.id:
                                        new_src_list.append([transition, state])

            elif elem[0] == "destination":
                for transition in xml_transition_list:
                    if elem[1] == transition.name or elem[1] == transition.alias:
                        for state in xml_state_list:
                            if elem[2] == state.name or elem[2] == state.alias:
                                if not isinstance(state.type, datamodel.BaseType):
                                    if 'ENTRY' in state.type.name:
                                        Logger.set_error(__name__,
                                                        f"{elem[2]} is typed as ENTRY state, it cannot be "
                                                        f"put as destination's transition (not added)")
                                else:
                                    if transition.destination != state.id:
                                        new_dest_list.append([transition, state])

    src_dest_lists = [new_src_list, new_dest_list]
    update = add_src_dest(src_dest_lists, output_xml)

    return update


def add_src_dest(src_dest_lists, output_xml):
    """
    Check if input lists are not empty, write in xml for each list and return update list if some
    updates has been made

        Parameters:
            src_dest_lists ([Transition, State(Source)],[Transition, State(Destination)]) :
            Transition object and Source/Destination
            output_xml (GenerateXML object) : XML's file object

        Returns:
            update_list ([0/1]) : Add 1 to list if any update, otherwise 0 is added
    """
    if any(src_dest_lists):
        new_src_list = src_dest_lists[0]
        new_dest_list = src_dest_lists[1]
        if new_src_list:
            output_xml.write_source(new_src_list)
            # Warn the user once writtent and added within xml
            for source in new_src_list:
                source[0].set_source(source[1].id)
                Logger.set_info(__name__,
                               f"{source[1].name} source for {source[0].name}")

        if new_dest_list:
            output_xml.write_destination(new_dest_list)
            # Warn the user once writtent and added within xml
            for destination in new_dest_list:
                destination[0].set_destination(destination[1].id)
                Logger.set_info(__name__,
                               f"{destination[1].name} destination for {destination[0].name}")
        return 1

    return 0


def check_add_exposes(exposes_str_list, xml_fun_elem_list, xml_fun_inter_list, xml_data_list,
                      output_xml):
    """
    Check and get all "Fun_elem exposes Fun_inter" strings, if Fun_inter is not exposed yet
    (or parentality relationship) => add it to Fun_elem object and as exposedInterface within xml.
    Args:
        exposes_str_list ([strings]): list of strings
        xml_fun_elem_list ([Fun Elem]) : Functional Element list from xml parsing
        xml_fun_inter_list ([FunctionalInterface]) : FunctionalInterface list from xml parsing
        xml_data_list ([Data]) : Data list from xml parsing
        output_xml (GenerateXML object) : XML's file object

    Returns:
        [0/1] : if update has been made
    """
    output = False
    for exposes_str in exposes_str_list:
        fun_elem = check_get_object(exposes_str[0], **{'xml_fun_elem_list': xml_fun_elem_list})
        fun_inter = check_get_object(exposes_str[1], **{'xml_fun_inter_list': xml_fun_inter_list})

        check_print_wrong_pair_object((exposes_str[0], fun_elem, 'Functional Element'),
                                      (exposes_str[1], fun_inter, 'Functional Interface'),
                                      'exposes')
        if fun_elem and fun_inter:
            check_rule = check_fun_elem_inter_families(fun_elem, fun_inter, xml_fun_elem_list)
            if fun_inter.id not in fun_elem.exposed_interface_list and check_rule:
                output = True
                fun_elem.add_exposed_interface(fun_inter.id)
                output_xml.write_exposed_interface([[fun_elem, fun_inter]])
                Logger.set_info(__name__,
                               f"{fun_elem.name} exposes {fun_inter.name}")

    if output:
        return 1

    return 0


def check_fun_elem_inter_families(fun_elem, fun_inter, xml_fun_elem_list):
    """A fun elem can expose an interface if already allocated to a parent/child and
    an interface can only be exposed by 2 families of fun_elem"""
    check = True
    exposed_fun_elem_list = set()
    for xml_fun_elem in xml_fun_elem_list:
        if any(s == fun_inter.id for s in xml_fun_elem.exposed_interface_list) and \
                xml_fun_elem != fun_elem:
            exposed_fun_elem_list.add(xml_fun_elem)

    if not exposed_fun_elem_list:
        return check

    opposite_fun_elem_list = []
    for elem in exposed_fun_elem_list:
        if check_parentality(elem, fun_elem) or \
                check_parentality(fun_elem, elem):
            return check
        else:
            opposite_fun_elem_list.append(elem)

    for idx in range(0, len(opposite_fun_elem_list)-1):
        if not check_parentality(
                opposite_fun_elem_list[idx], opposite_fun_elem_list[idx+1]) and \
                not check_parentality(opposite_fun_elem_list[idx+1], opposite_fun_elem_list[idx]):
            check = False
            return check

    return check


def check_print_wrong_pair_object(object_a, object_b, relationship_type):
    """
    Prints specific user messages for wrong object(s) pair (Object_a, Object_b) relationship

    Args:
        object_a: (input_string, object_or_none, object_type_string)
        e.g. (exposes_str[0], fun_elem, 'Functional Element')
        object_b: (exposes_str[1], fun_inter, 'Functional Interface')
        relationship_type: e.g. 'exposes'

    """
    if object_a[1] == object_b[1] is None:
        Logger.set_error(__name__,
                        f"{object_a[0]} and {object_b[0]} do not exist, choose valid names/aliases for: "
                        f"'{object_a[2]}' {relationship_type} "
                        f"'{object_b[2]}'")
    elif object_a[1] is None or object_b[1] is None:
        if object_a[1] is None and object_b[1]:
            Logger.set_error(__name__,
                            f"{object_a[0]} does not exist, choose a valid name/alias for: "
                            f"'{object_a[2]}' {relationship_type} "
                            f"{object_b[1].name}")
        elif object_b[1] is None and object_a[1]:
            Logger.set_error(__name__,
                            f"{object_b[0]} does not exist, choose a valid name/alias for: "
                            f"{object_a[1].name} {relationship_type} "
                            f"'{object_b[2]}'")



"""An example magic"""
__version__ = '0.0.1'
from . import command_parser
from .jarvis import MagicJarvis, greet_user, clean_folders
from tools import MagicTools, Config
from plantuml_adapter import PlantUmlGen


def load_ipython_extension(ipython):
    """Entry point from notebook/ipython when executing %(re)load_ext jarvis"""
    clean_folders()
    Config.read()
    generator = PlantUmlGen()
    parser = command_parser.CmdParser(generator)
    jarvis4se = MagicJarvis(ipython, parser)
    tools = MagicTools(ipython, generator)
    ipython.register_magics(jarvis4se, tools)
    greet_user()



from setuptools import setup

setup(name='plantuml_adapter',
version='0.1',
description='MTC PLANTUML adapter',
url='',
author='',
author_email='',
license='',
packages=['plantuml_adapter'],
zip_safe=False)


"""Module that contains tests for context diagrams"""
from conftest import get_jarvis4se, remove_xml_file
import plantuml_adapter

jarvis4se = get_jarvis4se()


def test_simple_function_context(mocker):
    """Notebook equivalent:
     %%jarvis
     with simple_function_context
     F1 is a function
     show context F1

     """
    spy = mocker.spy(plantuml_adapter, "get_function_diagrams")
    file_name = "simple_function_context"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "show context F1\n")
    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = 'object "F1" as f1 <<Function>>\n'

    remove_xml_file(file_name)

    assert expected in result
    assert len(result) - len(expected) == len("\'id: xxxxxxxxxx\n")


def test_simple_function_context_in_out(mocker):
    """Notebook equivalent:
     %%jarvis
     with simple_function_in_out
     F1 is a function
     a is a data
     b is a data
     =============================
     %%jarvis
     with simple_function_in_out
     F1 produces b
     =============================
     %%jarvis
     with simple_function_in_out
     F1 consumes a
     =============================
     %%jarvis
     with simple_function_in_out
     show context F1

     """
    spy = mocker.spy(plantuml_adapter, "get_function_diagrams")
    file_name = "simple_function_in_out"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "a is a data\n"
                         "b is a data\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 produces b\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 consumes a\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "show context F1\n")
    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = 'object "F1" as f1 <<Function>>\n' \
               'circle f1_i\n' \
               'circle f1_o\n' \
               'f1_i --> f1 : a\n' \
               'f1 --> f1_o  : b\n'

    remove_xml_file(file_name)

    assert expected in result
    assert len(result) - len(expected) == len("\'id: xxxxxxxxxx\n")


def test_function_context_with_attribute(mocker):
    """Notebook equivalent:
     %%jarvis
     with test_function_with_attribute
     F1 is a function
     ========================================
     %%jarvis
     with test_function_with_attribute
     A is an attribute
     C is an attribute
     ========================================
     %%jarvis
     with test_function_with_attribute
     The A of F1 is 4,2
     The C of F1 is pink
     ========================================
     %%jarvis
     with test_function_with_attribute
     show context F1
     """
    spy = mocker.spy(plantuml_adapter, "get_function_diagrams")
    file_name = "test_function_with_attribute"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "A is an attribute\n"
                         "C is an attribute\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "The A of F1 is 4,2\n"
                         "The C of F1 is pink\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "show context F1\n")
    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['object "F1" as f1 <<Function>> {\n', 'A = 4,2\n', 'C = pink\n', '}\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == len("\'id: xxxxxxxxxx\n")


def test_fun_elem_context_with_attribute(mocker):
    """Notebook equivalent:
     %%jarvis
     with fun_elem_context_with_attribute
     F1 is a function
     Fun elem is a functional element
     F1 is allocated to Fun elem
     ========================================
     %%jarvis
     with fun_elem_context_with_attribute
     A is an attribute
     B is an attribute. C is an attribute
     ========================================
     %%jarvis
     with fun_elem_context_with_attribute
     The A of F1 is 4,2
     The C of F1 is pink
     The B of Fun elem is 8,5.
     The A of Fun elem is 100
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_context_diagram")
    file_name = "fun_elem_context_with_attribute"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "Fun elem is a functional element\n"
                         "F1 is allocated to Fun elem")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "A is an attribute\n"
                         "B is an attribute. C is an attribute\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "The A of F1 is 4,2\n"
                         "The C of F1 is pink\n"
                         "The B of Fun elem is 8,5.\n"
                         "The A of Fun elem is 100\n")
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "show context Fun elem\n")
    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "Fun elem" as fun_elem <<Functional element>>{\n',
                'object "F1" as f1 <<Function>> {\n',
                'A = 4,2\n',
                'C = pink\n',
                '}\n}\n',
                'note bottom of fun_elem\n',
                'A = 100\n',
                'B = 8,5\n',
                'end note\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 2*len("\'id: xxxxxxxxxx\n")


def test_function_context_with_grandkids(mocker, function_grandkids_cell):
    """See Issue #31, Notebook equivalent:
    %%jarvis
    with function_context_with_grandkids
    F1 is a function
    F1a is a function
    F1a1 is a function
    F1 is composed of F1a
    F1a is composed of F1a1
    a is a data
    F1a produces a
    b is a data
    F1a consumes b
    c is a data
    F1a1 produces c
    d is a data
    F1a1 consumes d

    show context F1
     """
    spy = mocker.spy(plantuml_adapter, "get_function_diagrams")
    file_name = "function_context_with_grandkids"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         f"{function_grandkids_cell}\n"
                         f"show context F1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['object "F1" as f1 <<Function>>\n',
                'circle f1_i\n',
                'circle f1_o\n',
                'f1_i --> f1 : ',
                'b', '\\n', 'd', '\n',
                'f1 --> f1_o  : ',
                'c', '\\n', 'a', '\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == len("\'id: xxxxxxxxxx\n")


def test_fun_elem_context_with_interfaces(mocker):
    """Notebook equivalent:
    %%jarvis
    with fun_elem_context_with_interfaces
    F1 is a function
    F2 is a function
    A is a data
    B is a data
    C is a data
    F1 produces A
    F2 consumes A
    F2 produces B
    F1 consumes B
    F1 produces C
    Fun_elem_1 is a functional element
    Fun_elem_2 is a functional element
    Fun_inter_1 is a functional interface
    Fun_inter_2 is a functional interface
    Fun_elem_1 allocates F1
    Fun_elem_2 allocates F2
    Fun_inter_1 allocates A
    Fun_inter_2 allocates C
    Fun_elem_1 exposes Fun_inter_1
    Fun_elem_1 exposes Fun_inter_2
    Fun_elem_2 exposes Fun_inter_1

    show context Fun_elem_1
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_context_diagram")
    file_name = "fun_elem_context_with_interfaces"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "A is a data\n"
                         "B is a data\n"
                         "C is a data\n"
                         "F1 produces A\n"
                         "F2 consumes A\n"
                         "F2 produces B\n"
                         "F1 consumes B\n"
                         "F1 produces C\n"
                         "Fun_elem_1 is a functional element\n"
                         "Fun_elem_2 is a functional element\n"
                         "Fun_inter_1 is a functional interface\n"
                         "Fun_inter_2 is a functional interface\n"
                         "Fun_elem_1 allocates F1\n"
                         "Fun_elem_2 allocates F2\n"
                         "Fun_inter_1 allocates A\n"
                         "Fun_inter_2 allocates C\n"
                         "Fun_elem_1 exposes Fun_inter_1\n"
                         "Fun_elem_1 exposes Fun_inter_2\n"
                         "Fun_elem_2 exposes Fun_inter_1\n"
                         "\n"
                         "show context Fun_elem_1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "Fun_elem_2" as fun_elem_2 <<Functional element>>{\n',
                'object "F2" as f2 <<Function>>\n',
                '}\n',
                'component "Fun_elem_1" as fun_elem_1 <<Functional element>>{\n',
                'object "F1" as f1 <<Function>>\n',
                '}\n',
                'circle f1_o\n',
                'f1 --> f1_o  : C\n'
                'f2 #--> f1 : B\n',
                'fun_elem_1', ' -- ', 'fun_elem_2 ', ': fun_inter_1\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 4*len("\'id: xxxxxxxxxx\n")


def test_fun_elem_context_interface_with_no_flow(mocker):
    """ See issue #39, Notebook equivalent:
    %%jarvis
    with fun_elem_context_interface_with_no_flow
    E is a functional element
    E1 is a functional element
    I_E_E1 is a functional interface
    E exposes I_E_E1
    E1 exposes I_E_E1

    show context E1

    show context Fun_elem_1
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_context_diagram")
    file_name = "fun_elem_context_interface_with_no_flow"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "E is a functional element\n"
                         "E1 is a functional element\n"
                         "I_E_E1 is a functional interface\n"
                         "E exposes I_E_E1\n"
                         "E1 exposes I_E_E1\n"
                         "\n"
                         "show context E1\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E" as e <<Functional element>>{\n',
                '}\n',
                'component "E1" as e1 <<Functional element>>{\n',
                '}\n',
                'e1', ' -- ', 'e ', ': i_e_e1\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 2*len("\'id: xxxxxxxxxx\n")


def test_fun_elem_context_interface_not_exposed(mocker):
    """ See issue #38, Notebook equivalent:
    %%jarvis
    with fun_elem_context_interface_not_exposed
    F is a function
    F1 is a function
    F2 is a function
    a is a data
    F produces a
    F1 consumes a
    F2 consumes a
    b is a data
    F produces b
    F2 consumes b

    E is a functional element
    E allocates F
    E1 is a functional element
    E1 allocates F1
    E2 is a functional element
    E2 allocates F2

    I_E_E1 is a functional interface
    I_E_E1 allocates a
    E exposes I_E_E1
    E1 exposes I_E_E1

    show context E2
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_context_diagram")
    file_name = "fun_elem_context_interface_not_exposed"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F is a function\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "a is a data\n"
                         "F produces a\n"
                         "F1 consumes a\n"
                         "F2 consumes a\n"
                         "b is a data\n"
                         "F produces b\n"
                         "F2 consumes b\n"
                         "\n"
                         "E is a functional element\n"
                         "E allocates F\n"
                         "E1 is a functional element\n"
                         "E1 allocates F1\n"
                         "E2 is a functional element\n"
                         "E2 allocates F2\n"
                         "\n"
                         "I_E_E1 is a functional interface\n"
                         "I_E_E1 allocates a\n"
                         "E exposes I_E_E1\n"
                         "E1 exposes I_E_E1\n"
                         "\n"
                         "show context E2\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E" as e <<Functional element>>{\n',
                'object "F" as f <<Function>>\n',
                '}\n',
                'component "E2" as e2 <<Functional element>>{\n',
                'object "F2" as f2 <<Function>>\n',
                '}\n',
                'f #--> f2 : ',
                'a', '\\n', 'b', '\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 4*len("\'id: xxxxxxxxxx\n")


# Add childs to external fun_elem or create new test
def test_fun_elem_context_interface_with_child(mocker):
    """ See issue #44, Notebook equivalent:
    %%jarvis
    with fun_elem_context_interface_with_child
    F is a function
    F1 is a function
    a is a data
    F produces a
    F1 consumes a
    E is a functional element
    E1 is a functional element
    E allocates F
    E1 allocates F1
    I_E_E1 is a functional interface
    E exposes I_E_E1
    E1 exposes I_E_E1
    I_E_E1 allocates a

    E11 is a functional element
    E11 composes E
    E11 allocates F
    E11 exposes I_E_E1

    show context E
     """
    spy = mocker.spy(plantuml_adapter, "get_fun_elem_context_diagram")
    file_name = "fun_elem_context_interface_with_child"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F is a function\n"
                         "F1 is a function\n"
                         "a is a data\n"
                         "F produces a\n"
                         "F1 consumes a\n"
                         "E is a functional element\n"
                         "E1 is a functional element\n"
                         "E allocates F\n"
                         "E1 allocates F1\n"
                         "I_E_E1 is a functional interface\n"
                         "E exposes I_E_E1\n"
                         "E1 exposes I_E_E1\n"
                         "I_E_E1 allocates a\n"
                         "\n"
                         "E11 is a functional element\n"
                         "E11 composes E\n"
                         "E11 allocates F\n"
                         "E11 exposes I_E_E1\n"
                         "\n"
                         "show context E\n")

    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    expected = ['component "E" as e <<Functional element>>{\n',
                'object "F" as f <<Function>>\n',
                '}\n',
                'component "E1" as e1 <<Functional element>>{\n',
                'object "F1" as f1 <<Function>>\n',
                '}\n',
                'e1', ' -- ', 'e ', ': i_e_e1\n']

    remove_xml_file(file_name)

    assert all(i in result for i in expected)
    assert len(result) - len(''.join(expected)) == 4*len("\'id: xxxxxxxxxx\n")



from setuptools import setup


setup(name='datamodel',
      version='0.1',
      description='3SE datamodel',
      url='',
      author='',
      author_email='',
      license='',
      packages=['datamodel'],
      zip_safe=False)



""" @package tools
Tooling module

Defines the following classes that provides tooling features used by Jarvis:
- @ref MagicTools which defines magic line and cell for IPython interface
- @ref Logger which defines the logging mechanism
- @ref Config which defines the configuration mechanism

Provides the following utilities:
- @ref get_hyperlink to covert file path into HTML link

Provides the Jarvis console : @ref main
"""

from .util import get_hyperlink
from .magic_tools import MagicTools
from .config import Config
from .logger import Logger



# Libraries
from datetime import datetime
import os


# Modules
from . import Config


class Logger:
    """@ingroup tools
    @anchor Logger
    Jarvis logger
    """

    @classmethod
    def set_debug(cls, module_name, msg):
        """Format DEBUG message
        @param[in] module_name module name responsible for the message
        @param[in] msg message to be formatted
        @return None
        """
        print("[DEBUG] " + msg)
        if Config.is_log_file:
            cls.write("DEBUG", module_name, msg)

    @classmethod
    def set_info(cls, module_name, msg):
        """Format INFO message
        @param[in] module_name module name responsible for the message
        @param[in] msg message to be formatted
        @return None
        """
        print(msg)
        if Config.is_log_file:
            cls.write("INFO", module_name, msg)

    @classmethod
    def set_warning(cls, module_name, msg):
        """Format WARNING message
        @param[in] module_name module name responsible for the message
        @param[in] msg message to be formatted
        @return None
        """
        print("[WARNING] " + msg)
        if Config.is_log_file:
            cls.write("WARNING", module_name, msg)

    @classmethod
    def set_error(cls, module_name, msg):
        """Format ERROR message
        @param[in] module_name module name responsible for the message
        @param[in] msg message to be formatted
        @return None
        """
        print("[ERROR] " + msg)
        if Config.is_log_file:
            cls.write("ERROR", module_name, msg)

    @classmethod
    def write(cls, msg_type, module_name, msg):
        """Write message in the log file
        @param[in] msg_type message type
        @param[in] module_name module name responsible for the message
        @param[in] msg message to be written
        @return None
        """
        try:
            now = datetime.now()
            date = now.strftime("%Y-%m-%d")
            time = now.strftime("%H:%M:%S")

            if not os.path.isdir("log"):
                os.makedirs("log")

            with open("log/log_" + date + ".log", "a+", newline='', encoding='utf-8') as file_writer:
                file_writer.write(date + ";"
                                  + time + ";"
                                  + msg_type + ";"
                                  + module_name + ";"
                                  + msg + "\n")
        except EnvironmentError as ex:
            print(f"[ERROR] Unable to write log file in {os.getcwd()}\\log: {ex}")



"""Module that contains tests for State()"""
from conftest import get_jarvis4se, remove_xml_file
import plantuml_adapter

jarvis4se = get_jarvis4se()


def test_state_entry_exit_chain(mocker, state_exit_entry_chain_output_diagram):
    """Notebook equivalent:
     %%jarvis
     with state_entry_exit_chain
    EXIT_TOTO extends state
    ENTRY state extends state
    S1 is a EXIT_TOTO
    S2 is a ENTRY state
    show chain S1, S2
     """
    spy = mocker.spy(plantuml_adapter, "get_state_machine_diagram")
    file_name = "state_entry_exit_chain"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "EXIT_TOTO extends state\n"
                         "ENTRY state extends state\n"
                         "S1 is a EXIT_TOTO\n"
                         "S2 is a ENTRY state\n"
                         "show chain S1, S2\n")
    # result = plantuml text without "@startuml ... @enduml" tags
    result = spy.spy_return
    assert all(i in result for i in state_exit_entry_chain_output_diagram)
    assert len(result) - len(''.join(state_exit_entry_chain_output_diagram)) == 2*len("\'id: xxxxxxxxxx\n")

    remove_xml_file(file_name)


class Point:
    """@ingroup datamodel
    @anchor Point
    Basic type for diagram coordinates
    """
    def __init__(self, p_x='', p_y=''):
        """
        @var x
        X coordinate

        @var y
        Y coordinate
        """

        self.x = p_x
        self.y = p_y

    def set_x(self, p_x):
        """Set X coordinate
        @param[in] self this class instance
        @param[in] p_x X coordinate value
        @return None
        """
        self.x = p_x

    def set_y(self, p_y):
        """Set Y coordinate
        @param[in] self this class instance
        @param[in] p_y Y coordinate value
        @return None
        """
        self.y = p_y


class EndPoint:
    """@ingroup datamodel
    @anchor EndPoint
    Basic type for diagram object defined as a set of points
    """
    
    def __init__(self, p_name=''):
        """
        @var name
        unique name
        
        @var point_list
        point list
        """
        self.name = p_name
        self.point_list = set()

    def set_name(self, p_name):
        """Set unique name
        @param[in] self this class instance
        @param[in] p_name unique name
        @return None
        """
        self.name = p_name

    def add_point(self, p_point):
        """Add point to point_list
        @param[in] self this class instance
        @param[in] p_point point
        @return None
        """
        self.point_list.add(p_point)




"""Module that contains tests to check that xml file has been written"""
import os
from pathlib import Path

from conftest import get_jarvis4se, remove_xml_file
from xml_adapter import XmlParser3SE
from datamodel import BaseType

jarvis4se = get_jarvis4se()
xml_parser = XmlParser3SE()


def test_generate_xml_file_template():
    """Notebook equivalent:
     %%jarvis
     with generate_xml_file_template

     """
    file_name = "generate_xml_file_template"
    jarvis4se.jarvis("", f"with {file_name}\n")
    path = Path(os.path.join("./", file_name + ".xml"))
    with path as file:
        read_xml = file.read_text(encoding="utf-8")
        base_xml = "<?xml version='1.0' encoding='UTF-8'?>\n" \
                   "<systemAnalysis>\n" \
                   "  <funcArch>\n" \
                   "    <functionList/>\n" \
                   "    <dataList/>\n" \
                   "    <stateList/>\n" \
                   "    <transitionList/>\n" \
                   "    <functionalElementList/>\n" \
                   "    <functionalInterfaceList/>\n" \
                   "  </funcArch>\n" \
                   "  <phyArch>\n" \
                   "    <physicalElementList/>\n" \
                   "    <physicalInterfaceList/>\n" \
                   "  </phyArch>\n" \
                   "  <viewPoint>\n" \
                   "    <viewList/>\n" \
                   "    <attributeList/>\n" \
                   "    <typeList/>\n" \
                   "  </viewPoint>\n" \
                   "</systemAnalysis>\n"
        assert base_xml in read_xml
    remove_xml_file(file_name)


def test_simple_function_within_xml():
    """Notebook equivalent:
     %%jarvis
     with simple_function_within_xml
     F1 is a function

     """
    file_name = "simple_function_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function\n")

    function_list = xml_parser.parse_xml(file_name + ".xml")['xml_function_list']

    remove_xml_file(file_name)

    assert len(function_list) == 1
    assert [fun.name == "F1" for fun in function_list]


def test_described_attribute_within_xml(attribute_cell):
    """Same as test_described_attribute_input() within test_input_cell.py, but here we are
    verifying that attributes are written correctly within xml:
     %%jarvis
     with described_attribute_within_xml
     F1 is a function
     Fun elem is a functional element
     ========================================
     %%jarvis
     with described_attribute_within_xml
     A is an attribute
     B is an attribute. C is an attribute
     ========================================
     %%jarvis
     with described_attribute_within_xml
     The A of F1 is 4,2
     The C of F1 is pink
     The B of Fun elem is 8,5.
     The A of Fun elem is 100

     """
    file_name = "described_attribute_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[1]}")
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[2]}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    expected = {('A', 'F1', '4,2'), ('B', 'Fun elem', '8,5'),
                ('C', 'F1', 'pink'), ('A', 'Fun elem', '100')}
    # xml_adapter.parse_xml() returns mainly set(), so the order can change
    # thus we have to compare it with a set also
    result = set()
    assert len(obj_dict['xml_attribute_list']) == 3
    for attribute in obj_dict['xml_attribute_list']:
        for item in attribute.described_item_list:
            for function in obj_dict['xml_function_list']:
                if item[0] == function.id:
                    result.add((attribute.name, function.name, item[1]))
            for fun_elem in obj_dict['xml_fun_elem_list']:
                if item[0] == fun_elem.id:
                    result.add((attribute.name, fun_elem.name, item[1]))

    remove_xml_file(file_name)

    assert expected == result


def test_set_attribute_type_within_xml():
    """Tests that attribute types are written correctly within xml, notebook equivalent:
     %%jarvis
     with set_attribute_type_within_xml
     A is an attribute
     B is an attribute.
     attribute type A extends attribute
     attribute type B extends attribute
     The type of A is attribute type A.
     The type of B is attribute type B

     """
    file_name = "set_attribute_type_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "A is an attribute\n"
                         "B is an attribute.\n"
                         "attribute type A extends attribute\n"
                         "attribute type B extends attribute\n"
                         "The type of A is attribute type A.\n"
                         "The type of B is attribute type B\n")

    attribute_list = xml_parser.parse_xml(file_name + ".xml")['xml_attribute_list']
    expected = {('A', 'attribute type A'), ('B', 'attribute type B')}
    # xml_adapter.parse_xml() returns mainly set(), so the order can change
    # thus we have to compare it with a set also
    result = set()
    assert len(attribute_list) == 2
    for attribute in attribute_list:
        result.add((attribute.name, attribute.type.name))

    remove_xml_file(file_name)

    assert expected == result


def test_set_allocated_item_to_view_within_xml(allocation_item_cell):
    """Relative to Issue #9 to add new allocated item to a view(i.e. filter) by verifying than
    it's written within xml. Notebook equivalent:
    %%jarvis
    with set_allocated_item_to_view_within_xml
    F1 is a function
    F2 with a long name is a function. The alias of F2 with a long name is F2.
    F3 is a function
    F4 is a function
    a is a data
    Fun_elem is a functional element
    ========================================
    %%jarvis
    with set_allocated_item_to_view_within_xml
    under toto
    consider F1. consider toto. consider a, Fun_elem
    consider tata.
    consider F1, F2, F3, F4
    """
    file_name = "set_allocated_item_to_view_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{allocation_item_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{allocation_item_cell[1]}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    expected = {'F1', 'F2 with a long name', 'F3', 'F4', 'a', 'Fun_elem'}
    # xml_adapter.parse_xml() returns mainly set(), so the order can change
    # thus we have to compare it with a set also
    result = set()
    assert len(obj_dict['xml_view_list']) == 1
    assert "test_view" in {i.name for i in obj_dict['xml_view_list']}
    for item in next(iter(obj_dict['xml_view_list'])).allocated_item_list:
        for fun in obj_dict['xml_function_list']:
            if item == fun.id:
                result.add(fun.name)

        for fun_elem in obj_dict['xml_fun_elem_list']:
            if item == fun_elem.id:
                result.add(fun_elem.name)

        for data in obj_dict['xml_data_list']:
            if item == data.id:
                result.add(data.name)

    remove_xml_file(file_name)

    assert expected == result


def test_function_with_grandkids_within_xml(function_grandkids_cell):
    """See Issue #31, Notebook equivalent:
    %%jarvis
    with function_with_grandkids_within_xml
    F1 is a function
    F1a is a function
    F1a1 is a function
    F1 is composed of F1a
    F1a is composed of F1a1
    a is a data
    F1a produces a
    b is a data
    F1a consumes b
    c is a data
    F1a1 produces c
    d is a data
    F1a1 consumes d
    """
    file_name = "function_with_grandkids_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{function_grandkids_cell}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    expected_cons = {('b', 'F1a'), ('d', 'F1'), ('b', 'F1'), ('d', 'F1a'), ('d', 'F1a1')}
    expected_prod = {('c', 'F1a1'), ('a', 'F1'), ('c', 'F1'), ('c', 'F1a'), ('a', 'F1a')}
    expected_child = {('F1', 'F1a'), ('F1a', 'F1a1')}
    # xml_adapter.parse_xml() returns mainly set(), so the order can change
    # thus we have to compare it with a set also
    result_cons = set()
    result_prod = set()
    result_child = set()
    assert len(obj_dict['xml_data_list']) == 4 and len(obj_dict['xml_function_list']) == 3
    assert (len(obj_dict['xml_consumer_function_list']) and
            len(obj_dict['xml_producer_function_list'])) == 5

    for cons in obj_dict['xml_consumer_function_list']:
        result_cons.add((cons[0], cons[1].name))
    for prod in obj_dict['xml_producer_function_list']:
        result_prod.add((prod[0], prod[1].name))
    for fun in obj_dict['xml_function_list']:
        if fun.child_list:
            for child in fun.child_list:
                result_child.add((fun.name, child.name))

    remove_xml_file(file_name)

    assert expected_cons == result_cons
    assert expected_prod == result_prod
    assert expected_child == result_child


def test_function_childs_cons_prod_within_xml(function_with_childs_cell):
    """See Issue #5, Notebook equivalent:
    %%jarvis
    with function_childs_cons_prod_within_xml
    F1 is a function
    F1a is a function
    F1b is a function
    F1c is a function
    F1d is a function
    F1e is a function
    F2 is a function
    F3 is a function

    F1 is composed of F1a
    F1 is composed of F1b
    F1 is composed of F1c
    F1 is composed of F1d
    F1 is composed of F1e

    a is a data
    F1 produces a
    F2 consumes a

    F1a produces a
    F1b consumes a

    b is a data
    F1c produces b
    F1d consumes b

    c is a data
    F3 produces c
    F1e consumes c
    """
    file_name = "function_childs_cons_prod_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         f"{function_with_childs_cell}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    expected_cons = {('a', 'F1b'), ('b', 'F1d'), ('a', 'F2'), ('c', 'F1e'), ('c', 'F1')}
    expected_prod = {('b', 'F1c'), ('c', 'F3'), ('a', 'F1a'), ('a', 'F1')}
    expected_child = {('F1', 'F1e'), ('F1', 'F1d'), ('F1', 'F1c'), ('F1', 'F1b'), ('F1', 'F1a')}
    # xml_adapter.parse_xml() returns mainly set(), so the order can change
    # thus we have to compare it with a set also
    result_cons = set()
    result_prod = set()
    result_child = set()
    assert len(obj_dict['xml_data_list']) == 3 and len(obj_dict['xml_function_list']) == 8
    assert len(obj_dict['xml_consumer_function_list']) == 5 and \
           len(obj_dict['xml_producer_function_list']) == 4

    for cons in obj_dict['xml_consumer_function_list']:
        result_cons.add((cons[0], cons[1].name))
    for prod in obj_dict['xml_producer_function_list']:
        result_prod.add((prod[0], prod[1].name))
    for fun in obj_dict['xml_function_list']:
        if fun.child_list:
            for child in fun.child_list:
                result_child.add((fun.name, child.name))

    remove_xml_file(file_name)

    assert expected_cons == result_cons
    assert expected_prod == result_prod
    assert expected_child == result_child


def test_functional_interface_within_xml():
    """Notebook equivalent:
    %%jarvis
    with functional_interface_within_xml
    Color is an attribute
    A is a data
    F1 is a function
    F2 is a function
    Fun_elem_1 is a functional element
    Fun_elem_2 is a functional element
    F1 produces A
    F2 consumes A
    Fun_elem_1 allocates F1
    Fun_elem_2 allocates F2
    Fun_inter is a functional interface.
    The type of Fun_inter is a_type
    The alias of Fun_inter is FI
    The Color of Fun_inter is pink
    Fun_elem_1 exposes Fun_inter
    Fun_elem_2 exposes Fun_inter
    Fun_inter allocates A.
    """
    file_name = "functional_interface_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "Color is an attribute\n"
                         "A is a data\n"
                         "F1 is a function\n"
                         "F2 is a function\n"
                         "Fun_elem_1 is a functional element\n"
                         "Fun_elem_2 is a functional element\n"
                         "F1 produces A\n"
                         "F2 consumes A\n"
                         "Fun_elem_1 allocates F1\n"
                         "Fun_elem_2 allocates F2\n"
                         "Fun_inter is a functional interface.\n"
                         "The type of Fun_inter is functional interface\n"
                         "The alias of Fun_inter is FI\n"
                         "The Color of Fun_inter is pink\n"
                         "Fun_elem_1 exposes Fun_inter\n"
                         "Fun_elem_2 exposes Fun_inter\n"
                         "Fun_inter allocates A.\n")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    assert (len(obj_dict['xml_data_list']) == len(obj_dict['xml_attribute_list']) ==
            len(obj_dict['xml_fun_inter_list'])) == 1
    data = obj_dict['xml_data_list'].pop()
    fun_inter = obj_dict['xml_fun_inter_list'].pop()
    attribute = obj_dict['xml_attribute_list'].pop()
    assert data.name == 'A'
    assert fun_inter.name == 'Fun_inter'
    assert fun_inter.alias == 'FI'
    assert fun_inter.type == BaseType['FUNCTIONAL_INTERFACE']
    assert attribute.name == 'Color'
    described_item = attribute.described_item_list.pop()
    assert described_item[0] == fun_inter.id and described_item[1] == 'pink'
    assert fun_inter.allocated_data_list.pop() == data.id

    remove_xml_file(file_name)


def test_fun_elem_exposes_interface_within_xml(fun_elem_exposing_cell):
    """Notebook equivalent:
    %%jarvis
    with fun_elem_exposes_interface_within_xml
    Fun_inter is a functional interface
    Fun_elem is a functional element
    Fun_elem_2 is a functional element
    Fun_elem_3 is a functional element
    Fun_elem_4 is a functional element
    Fun_elem_5 is a functional element
    Fun_elem_6 is a functional element
    Fun_elem_ext is a functional element
    Fun_elem is composed of Fun_elem_2
    Fun_elem_2 is composed of Fun_elem_3
    Fun_elem_3 is composed of Fun_elem_4
    Fun_elem_4 is composed of Fun_elem_5
    Fun_elem_5 is composed of Fun_elem_6
    Fun_elem exposes Fun_inter
    Fun_elem_6 exposes Fun_inter
    Fun_elem_ext exposes Fun_inter
    toto exposes Fun_inter
    tata exposes titi
    Fun_elem exposes coco
    """
    file_name = "fun_elem_exposes_interface_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{fun_elem_exposing_cell}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    expected_child = {('Fun_elem', 'Fun_elem_2'), ('Fun_elem_2', 'Fun_elem_3'),
                      ('Fun_elem_3', 'Fun_elem_4'), ('Fun_elem_4', 'Fun_elem_5'),
                      ('Fun_elem_5', 'Fun_elem_6')}
    expected_exposed = {('Fun_elem', 'Fun_inter'), ('Fun_elem_6', 'Fun_inter'),
                        ('Fun_elem_ext', 'Fun_inter')}

    assert len(obj_dict['xml_fun_inter_list']) == 1 and len(obj_dict['xml_fun_elem_list']) == 8
    fun_inter = obj_dict['xml_fun_inter_list'].pop()
    assert fun_inter.name == 'Fun_inter'

    result_exposed = set()
    result_child = set()
    for fun_elem in obj_dict['xml_fun_elem_list']:
        for child in fun_elem.child_list:
            result_child.add((fun_elem.name, child.name))
        if fun_inter.id in fun_elem.exposed_interface_list:
            result_exposed.add((fun_elem.name, fun_inter.name))

    assert expected_child == result_child
    assert expected_exposed == result_exposed

    remove_xml_file(file_name)


def test_type_within_xml(extends_and_set_type_cell):
    """ Issue #56 Notebook equivalent:
    %%jarvis
    with type_within_xml
    Safety interface extends functional interface
    The alias of Safety interface is sf
    ========================================
    %%jarvis
    sf_a extends sf
    sf_a_b extends sf_a
    final one extends sf_a_b
    Fun_inter is a functional interface
    The type of Fun_inter is final one
    """
    file_name = "type_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_set_type_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_set_type_cell[1]}")

    obj_dict = xml_parser.parse_xml(file_name + ".xml")

    assert len([x for x in obj_dict.values() if x]) == 2
    assert len(obj_dict['xml_type_list']) == 4
    assert len(obj_dict['xml_fun_inter_list']) == 1

    expected_type = {('sf_a', 'Safety interface'), ('sf_a_b', 'sf_a'),
                     ('Safety interface', 'Functional interface'), ('final one', 'sf_a_b')}
    captured_type = set()
    for type_elem in obj_dict['xml_type_list']:
        if type_elem.name == 'Safety interface':
            assert type_elem.alias == 'sf'
        if isinstance(type_elem.base, BaseType):
            base_type = str(type_elem.base)
        else:
            base_type = type_elem.base.name
        captured_type.add((type_elem.name, base_type))

    assert expected_type == captured_type
    assert obj_dict['xml_fun_inter_list'].pop().type.name == "final one"

    remove_xml_file(file_name)


def test_extends_and_create_object_within_xml(capsys, extends_and_create_object_cell):
    """ Issue #62 Notebook equivalent:
    %%jarvis
    with extends_and_create_object_within_xml
    "High level function" extends function
    "High high level function" extends "High level function"
    "High high high level function" extends "High high level function"
    3High is a "High high high level function"
    """
    file_name = "extends_and_create_object_within_xml"
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_create_object_cell}")
    
    obj_dict = xml_parser.parse_xml(file_name + ".xml")
    
    assert len([x for x in obj_dict.values() if x]) == 2
    assert len(obj_dict['xml_type_list']) == 3
    assert len(obj_dict['xml_function_list']) == 1

    expected_type = {('High level function', 'Function'),
                     ('High high level function', 'High level function'),
                     ('High high high level function', 'High high level function')}
    captured_type = set()
    for type_elem in obj_dict['xml_type_list']:
        print(type_elem, type_elem.base)
        if isinstance(type_elem.base, BaseType):
            base_type = str(type_elem.base)
        else:
            base_type = type_elem.base.name
        captured_type.add((type_elem.name, base_type))

    assert expected_type == captured_type
    assert obj_dict['xml_function_list'].pop().type.name == "High high high level function"

    remove_xml_file(file_name)



""" @package plantuml_adapter
Plantuml adapter module

Defines the different functions and classes used by Jarvis to call Plantuml as a diagram generator:
- function to construct the PlantUml text and url for the functions diagram: @ref get_function_diagrams
- function to construct the PlantUml text for the sequence diagrams: @ref get_sequence_diagram
- function to construct the PlantUml text and url for state machine diagrams:
@ref get_state_machine_diagram
- function to construct the PlantUml text for the functional element decomposition diagram:
@ref get_fun_elem_decomposition
- function to construct the PlantUml text and url for the context diagram for functional elements:
@ref get_fun_elem_context_diagram
- class to handle local PlantUml PicoWeb Server: @ref PlantUmlPicoServer
- class to encode PlantUml text and get server url as .svg : @ref PlantUmlGen
- class to encode PlantUml text for state diagram : @ref StateDiagram
- class to encode PlantUml text for sequence diagram : @ref SequenceDiagram
- class to encode PlantUml text for object diagram : @ref ObjDiagram
"""

from .plantuml_adapter import get_function_diagrams
from .plantuml_adapter import get_sequence_diagram
from .plantuml_adapter import get_state_machine_diagram
from .plantuml_adapter import get_fun_elem_decomposition
from .plantuml_adapter import get_fun_elem_context_diagram
from .util import PlantUmlPicoServer
from .util import PlantUmlGen



"""Modules containing tests to check jarvis ouput i.e. jupyter notebook output messages"""
from conftest import get_jarvis4se, remove_xml_file

jarvis4se = get_jarvis4se()


def test_attribute_declaration_input(capsys):
    """Notebook equivalent:
     %%jarvis
     with attribute_declaration_input
     A is an attribute
     B is an attribute. C is an attribute

     """
    file_name = "attribute_declaration_input"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "A is an attribute\n"
                         "B is an attribute. C is an attribute\n")

    captured = capsys.readouterr()
    expected = [f"Creating {file_name}.xml !\n",
                "C is an attribute\n",
                "A is an attribute\n",
                "B is an attribute\n",
                f"{file_name}.xml updated"]

    remove_xml_file(file_name)

    assert all(i in captured.out for i in expected)


def test_described_attribute_input(capsys, attribute_cell):
    """Notebook equivalent:
     %%jarvis
     with described_attribute_input
     F1 is a function
     Fun elem is a functional element
     ========================================
     %%jarvis
     with described_attribute_input
     A is an attribute
     B is an attribute. C is an attribute
     ========================================
     %%jarvis
     with described_attribute_input
     The A of F1 is 4,2
     The C of F1 is pink
     The B of Fun elem is 8,5.
     The A of Fun elem is 100

     """
    file_name = "described_attribute_input"
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[1]}")
    jarvis4se.jarvis("", f"with {file_name}\n{attribute_cell[2]}")

    captured = capsys.readouterr()
    expected = [f"{file_name}.xml parsed\n",
                "Attribute A for F1 with value 4,2\n",
                "Attribute C for F1 with value pink\n",
                "Attribute B for Fun elem with value 8,5\n",
                "Attribute A for Fun elem with value 100\n",
                f"{file_name}.xml updated\n"]
    # Get las part from capsys
    last_out = captured.out[-len(''.join(expected))-1:len(captured.out)]

    remove_xml_file(file_name)

    assert all(i in last_out for i in expected)


def test_set_object_type_alias_input(capsys):
    # Can be extend to test with source/destination
    """ In order to check Issue #21 causing bad regex match between type, alias and
    described attribute. Notebook equivalent:
    %%jarvis
    with set_object_type_alias_input
    F1 is a function.
    high level function extends function. The alias of F1 is f1. The type of f1 is high level function
    ========================================

    """
    file_name = "set_object_type_alias_input"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "F1 is a function.\n"
                         "high level function extends function. The alias of F1 is f1. "
                         "The type of f1 is high level function\n")

    captured = capsys.readouterr()
    expected = [f"Creating {file_name}.xml !\n",
                "high level function is a type extending Function\n"
                "F1 is a Function\n",
                "The alias for F1 is f1\n",
                "The type of F1 is high level function\n",
                f"{file_name}.xml updated\n"]

    remove_xml_file(file_name)

    assert all(i in captured.out for i in expected)
    assert len(captured.out) == len(''.join(expected))


def test_consider_object_input(capsys, allocation_item_cell):
    """ Relative to Issue #9 to add new allocated item to a view(i.e. filter). Notebook equivalent:
    %%jarvis
    with consider_object_input
    F1 is a function
    F2 with a long name is a function. The alias of F2 with a long name is F2
    F3 is a function
    F4 is a function
    a is a data
    Fun_elem is a functional element
    ========================================
    %%jarvis
    with consider_object_input
    under test_view
    consider F1. consider toto. consider a, Fun_elem
    consider tata.
    consider F1, F2, F3, F4
    """
    file_name = "consider_object_input"
    jarvis4se.jarvis("", f"with {file_name}\n{allocation_item_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{allocation_item_cell[1]}")

    captured = capsys.readouterr()

    expected = [f"{file_name}.xml parsed\n",
                "test_view is a view\n",
                "[WARNING] Object toto does not exist, available object types are : "
                "Functional Element, Function and Data\n",
                "[WARNING] Object tata does not exist, available object types are : "
                "Functional Element, Function and Data\n",
                "Function F1 is allocated to View test_view\n",
                "Data a is allocated to View test_view\n",
                "FunctionalElement Fun_elem is allocated to View test_view\n",
                "Function F2 with a long name is allocated to View test_view\n",
                "Function F3 is allocated to View test_view\n",
                "Function F4 is allocated to View test_view\n",
                f"{file_name}.xml updated\n"]

    # Get last part from capsys
    last_out = captured.out[-len(''.join(expected))-1:len(captured.out)]

    remove_xml_file(file_name)

    assert all(i in last_out for i in expected)


def test_functional_interface_input(capsys):
    """Notebook equivalent:
    %%jarvis
    with functional_interface_input
    Color is an attribute
    A is a data
    Fun_inter is a functional interface.
    The alias of Fun_inter is FI
    The Color of Fun_inter is pink
    Fun_inter allocates A.
    """
    file_name = "functional_interface_input"
    jarvis4se.jarvis("", f"with {file_name}\n"
                         "Color is an attribute\n"
                         "A is a data\n"
                         "Fun_inter is a functional interface.\n"
                         "The alias of Fun_inter is FI\n"
                         "The Color of Fun_inter is pink\n"
                         "Fun_inter allocates A.\n")

    captured = capsys.readouterr()
    expected = [f"Creating {file_name}.xml !\n",
                "A is a Data\n",
                "Fun_inter is a Functional interface\n",
                "Color is an attribute\n",
                "The alias for Fun_inter is FI\n",
                "[ERROR] Data A has no producer(s) nor consumer(s) allocated to functional "
                "elements exposing Fun_inter, A not allocated to Fun_inter\n",
                "Attribute Color for Fun_inter with value pink\n",
                f"{file_name}.xml updated\n"]

    remove_xml_file(file_name)

    assert len(captured.out) == len("".join(expected))
    assert all(i in captured.out for i in expected)


def test_fun_elem_exposes_interface_input(capsys, fun_elem_exposing_cell):
    """Notebook equivalent:
    %%jarvis
    with fun_elem_exposes_interface_input
    Fun_inter is a functional interface
    Fun_elem is a functional element
    Fun_elem_2 is a functional element
    Fun_elem_3 is a functional element
    Fun_elem_4 is a functional element
    Fun_elem_5 is a functional element
    Fun_elem_6 is a functional element
    Fun_elem_ext is a functional element
    Fun_elem_ext_2 is a functional element
    Fun_elem is composed of Fun_elem_2
    Fun_elem_2 is composed of Fun_elem_3
    Fun_elem_3 is composed of Fun_elem_4
    Fun_elem_4 is composed of Fun_elem_5
    Fun_elem_5 is composed of Fun_elem_6
    Fun_elem exposes Fun_inter
    Fun_elem_6 exposes Fun_inter
    Fun_elem_ext exposes Fun_inter
    Fun_elem_ext_2 exposes Fun_inter
    toto exposes Fun_inter
    tata exposes titi
    Fun_elem exposes coco
    """
    file_name = "fun_elem_exposes_interface_input"
    jarvis4se.jarvis("", f"with {file_name}\n{fun_elem_exposing_cell}")

    captured = capsys.readouterr()
    expected = ["Fun_elem exposes Fun_inter\n",
                "Fun_elem_6 exposes Fun_inter\n",
                "Fun_elem_ext exposes Fun_inter\n",
                "[ERROR] toto does not exist, choose a valid name/alias for: "
                "'Functional Element' exposes Fun_inter\n",
                "[ERROR] tata and titi do not exist, choose valid names/aliases for: "
                "'Functional Element' exposes 'Functional Interface'\n",
                "[ERROR] coco does not exist, choose a valid name/alias for: "
                "Fun_elem exposes 'Functional Interface'\n",
                f"{file_name}.xml updated\n"]
    # Get last part from capsys
    last_out = captured.out[-len(''.join(expected)):len(captured.out)]

    remove_xml_file(file_name)

    assert all(i in last_out for i in expected)


def test_extends_and_set_type_object_input(capsys, extends_and_set_type_cell):
    """ Issue #56 Notebook equivalent:
    %%jarvis
    with extends_and_set_type_object_input
    Safety interface extends functional interface
    The alias of Safety interface is sf
    ========================================
    %%jarvis
    with extends_and_set_type_object_input
    sf_a extends sf
    sf_a_b extends sf_a
    final one extends sf_a_b
    Fun_inter is a functional interface
    The type of Fun_inter is final one
    """
    file_name = "extends_and_set_type_object_input"
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_set_type_cell[0]}")
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_set_type_cell[1]}")

    captured = capsys.readouterr()
    expected = [f"Creating {file_name}.xml !\n",
                "Safety interface is a type extending Functional interface\n",
                "The alias for Safety interface is sf\n",
                f"{file_name}.xml updated\n"
                f"{file_name}.xml parsed\n",
                "Fun_inter is a Functional interface\n",
                "sf_a is a type extending Safety interface\n",
                "sf_a_b is a type extending sf_a\n",
                "final one is a type extending sf_a_b\n",
                "The type of Fun_inter is final one\n",
                f"{file_name}.xml updated\n"]

    remove_xml_file(file_name)

    assert all(i in captured.out for i in expected)


def test_extends_and_create_object_input(capsys, extends_and_create_object_cell):
    """ Issue #62 Notebook equivalent:
    %%jarvis
    with extends_and_create_object_input
    "High level function" extends function
    "High high level function" extends "High level function"
    "High high high level function" extends "High high level function"
    3High is a "High high high level function"
    """
    file_name = "extends_and_create_object_input"
    jarvis4se.jarvis("", f"with {file_name}\n{extends_and_create_object_cell}")

    captured = capsys.readouterr()
    expected = [f"Creating {file_name}.xml !\n",
                "High level function is a type extending Function\n",
                "High high level function is a type extending High level function\n",
                "High high high level function is a type extending High high level function\n"
                "3High is a High high high level function\n",
                f"{file_name}.xml updated\n"]

    remove_xml_file(file_name)

    assert all(i in captured.out for i in expected)
